#include "pr4f.h"
#include "aa_empirical.h"
FILE *fopene(char *filename, char *rw, char *err){
  FILE *fp;
  if((fp=fopen(filename,rw))==NULL){
    printf("%s\n",err);
    exit(1);
  }
  return(fp);
}
char get_param(int *k, char **arg, char *value){
  char opt='z';
  int i;
  
  for(i = *k; i >= 0; i--){
    if(arg[i][0]=='-'){
      opt=arg[i][1];
      if(arg[i][1]=='\0'){
	*value='\0'; 
      }
      else{
	if(arg[i][2]!='\0') sprintf(value,"%s",&arg[i][2]);
	if(arg[i][2]=='\0') sprintf(value,"%s",arg[i+1]);
      }
    }
    if(opt!='z'){
      *k=i-1;
      break;
    }
  }
  return(opt);
}
char ctl_paraml(FILE *fp, char *opt, char *value, int *is_option){
  /* return last char on line */
  char c;
  int i;

  *is_option=0;
  while(((c = getc(fp)) == ' ' || c == '\t' || c=='\r') && c != EOF && c != '\n');
  if(c == EOF || c == '\n') return(c);
  if(c == '*'){ /* comment */
    while((c = getc(fp)) != '\n' && c != EOF) ;
    return(c);
  } 

  /* option name */
  *is_option=1;
  for(i=0; c !=' ' && c != '\t' && c != '\r'; i++){
    opt[i]=c;
    c=getc(fp);
  }
  opt[i] = '\0';
  /* skip '=' */
  while((c=getc(fp))==' ' || c=='\t' || c == '=') ;
  for(i=0; c != ' ' && c != '\t' && c != '\r' && c != '\n' && c != EOF; i++){
    value[i]=c;
    c=getc(fp);
  }
  value[i]= '\0';
  while(c != '\n' && c != EOF) c=getc(fp);
  return(c);
}
int pmodel2smodel(int model, char *aaRmat, int nchar){
  if (nchar==4){
    if(model==0) return(0); /* JC */
    if(model==2) return(1); /* F81 */
    if(model==3) return(2); /* F84 */
    if(model==4) return(8); /* HKY */
    if(model==7) return(3); /* GTR */
  }
  if (nchar != 4){
    if(model==0) return(0); /* JC */
    if(model==1) return(1); /* F81 */
    if(model==8) return(3); /* GTR */
  }
  if (*aaRmat!='\0'){
    if(strcmp(aaRmat,"dayhoff.dat")==0) return(4); /* PAM */
    if(strcmp(aaRmat,"jones.dat")==0) return(5); /* JTT */
    if(strcmp(aaRmat,"wag.dat")==0) return(6); /* WAG */
    if(strcmp(aaRmat,"mtREV24.dat")==0) return(7); /* mtREV24 */
    if(strcmp(aaRmat,"lg.dat")==0) return(9); /* LG */
  }
  printf("No substitution model available\n");
  exit(1);
}
void seq_genc(int smodel, int ntaxa, double *utreec, int nsite, int *seq, 
	      int nrate, double alpha, int iseed, double *fr, double *Qk)
{
  /* IN:
   * smodel - 0:JC, 1:F81, 2:F84, 3:GTR, 4:PAM, 5:JTT, 6:WAG, 7:mtREV24, 8:HKY,
   *          9:LG
   * nrate, alpha - if alpha > 0, gamma model generation. If nrate>0, nrate rates
   * iseed - used to set seed
   * fr - frequencies if different than model frequencies
   *      if *fr < 0 for aa models, model-based frequencies used
   * Qk - rate matrix for GTR, ttratio, K or Kappa for F84, HKY, < 0 => ttratio; 
   * not used otherwise.
   *
   * OUT:
   * seq
   * infile to tmp_seq_genc.infile 
   * stderr to tmp_seq_genc.out
   * tree in tmp_seq_genc.intree
   *    all removed afterwards
   */
  FILE *fp;
  char (*names)[11],tstring[10000],cmdline[10000],tmpstr[1000],*seqc,*oseqc;
  char smodels[10][8]={"F84","F84","F84","GTR","PAM","JTT","WAG","MTREV","HKY",
		       "GENERAL"};
  int i,j,k,nchar=4,smodell,so,fso;
  double fro[20],W[400],Lam[20],R[400];

  seqc = (char *) malloc((size_t) nsite*ntaxa*sizeof(char));
 
  /* create seq-gen command and run */
  if (smodel == 4 || smodel == 5 || smodel == 6 || smodel == 7 || smodel == 9) 
    nchar = 20;
  names = (char (*)[11]) malloc((size_t) ntaxa*sizeof(*names));
  for(i = 0; i < ntaxa; i++) sprintf(names[i],"%i",i);
  /* pr_utreec(stdout,ntaxa,utreec); */
  ut2n(ntaxa,utreec,names,tstring);
/*   printf("%s\n",tstring); */
  fp=fopen("tmp_seq_genc.intree","w");
  fprintf(fp,"%s\n",tstring);
  fclose(fp);
  smodell=smodel;
  if(smodel==3 && nchar==20) smodell=9;
  sprintf(cmdline,"seq-gen -m%s -z%i -l%i",smodels[smodell], iseed, nsite);
  if (alpha > 0){
    sprintf(tmpstr," -a%f",alpha);
    strcat(cmdline,tmpstr);
    if(nrate > 0) sprintf(tmpstr," -g%i",nrate);
    strcat(cmdline,tmpstr);
  }
  /* printf("%s\n",cmdline); */
  if (smodel == 0 || smodel == 1 || smodel == 2 || smodel == 8){
    if (smodel == 0 || smodel == 1) sprintf(tmpstr," -t%f",K2tt(0,fr));
    if (smodel == 2){ 
      if (*Qk < 0)
	sprintf(tmpstr," -t%f",-(*Qk));
      else
	sprintf(tmpstr," -t%f",K2tt(*Qk,fr));
    }
    if (smodel == 8){ 
      if (*Qk < 0)
	sprintf(tmpstr," -t%f",-(*Qk));
      else
	sprintf(tmpstr," -t%f",kappa2tt(*Qk,fr));
    }
    strcat(cmdline,tmpstr);
  }
  if (*fr > 0){
    strcat(cmdline," -f");
    for(i = 0; i < nchar; i++){
      if(i==0) 
	sprintf(tmpstr,"%f",fr[i]);
      else
	sprintf(tmpstr,",%f",fr[i]);
      strcat(cmdline,tmpstr);
    }
  }
  if(smodel == 3){
    strcat(cmdline," -r");
    for(i = 0; i < nchar; i++)
      for(j = i+1; j < nchar; j++){
	if(i==0&&j==0)
	  sprintf(tmpstr,"%f",Qk[i+j*nchar]/fr[j]);
	else
	  sprintf(tmpstr,",%f",Qk[i+j*nchar]/fr[j]);
	strcat(cmdline,tmpstr);
      }
  }
  
  if(smodel==9){/* LG model not directly avail. use GTR */
    makeprotfreqs(fro,lgprobmat,lgeigmat);
    if (*fr > 0){
      memcpy(W,lgprobmat,(size_t) nchar*nchar*sizeof(double));
      memcpy(Lam,lgeigmat,(size_t) nchar*sizeof(double));
      chQ_freq(nchar,W,Lam,fr,fro);
      WLamfr2exchange(nchar,W,Lam,fr,R);
    }
    if (*fr < 0){
      strcat(cmdline," -f");
      for(i = 1; i < nchar; i++){
	if(i==0)
	  sprintf(tmpstr,"%f",fro[i]);
	else
	  sprintf(tmpstr,",%f",fro[i]);
	strcat(cmdline,tmpstr);
      }
      WLamfr2exchange(nchar,lgprobmat,lgeigmat,fro,R);
    }
    strcat(cmdline," -r");
    for(i = 0; i < nchar; i++)
      for(j = i+1; j < nchar; j++){
	if(i==0&&j==1)
	  sprintf(tmpstr,"%.2f",R[i+j*nchar]);
	else
	  sprintf(tmpstr,",%.2f",R[i+j*nchar]);
	strcat(cmdline,tmpstr);
      }
  }

  strcat(cmdline," < tmp_seq_genc.intree > tmp_seq_genc.infile 2> tmp_seq_genc.outfile");
  /* printf("%s\n",cmdline); */
  so=system(cmdline);
  
  /* read in the sequence */
  fp=fopen("tmp_seq_genc.infile","r");
  fso=fscanf(fp,"%i %i",&ntaxa,&nsite);
  rinterleavef_nolim(fp,&ntaxa,&nsite,names,seqc);
  fclose(fp);
  remove("tmp_seq_genc.infile");
  remove("tmp_seq_genc.intree");
  remove("tmp_seq_genc.outfile");

  /* sort so that integer names ordered low to high */
  oseqc = (char *) malloc(nsite*ntaxa*sizeof(char));
  for(j = 0; j < ntaxa; j++)
    for(k = 0; k < ntaxa; k++)
      if (atoi(names[k]) == j)
	for(i = 0; i < nsite; i++) oseqc[i+j*nsite]=seqc[i+k*nsite];
  if (nchar == 20)
    for(i = 0; i < nsite*ntaxa; i++) seq[i]=l2ip(oseqc[i]);
  if (nchar == 4)
    for(i = 0; i < nsite*ntaxa; i++) seq[i]=l2i(oseqc[i]);
  free(names); free(seqc); free(oseqc);
  return;
}
void pr_splits(FILE *outfile, int numsp, int nspl, int *splits){
  int split,i;
  for(split = 0; split < nspl; split++){
    for(i = 0; i < numsp; i++) fprintf(outfile, "%2i ",splits[i+split*numsp]);
    fprintf(outfile, "\n");
  }
}
void pr_usplits(FILE *outfile, int nspl, int numsp, int *usplits, int *count){
  int i,j;
  for(i = 0; i < nspl; i++){
    printf("%4i: ", i);
    for(j = 0; j < numsp; j++) fprintf(outfile, "%i ",usplits[j+i*numsp]);
    fprintf(outfile, "%i \n",count[i]);
  }
}
void split_leading1(int numsp, int *split){
  int i;
  if (split[0] == 1) return;
  for(i = 0; i < numsp; i++) split[i] = (split[i] == 1)?0:1;
}
void utreec2splits(int numsp, double *utreec, int *splits){
  int i,r,l,split;
  /* ith column gives split for numsp+ith branch */
  for(split = 0; split < numsp-3; split++){
    r = (int) utreec[split*4]; l = (int) utreec[1+split*4];
    if (r < numsp){
      for(i = 0; i < numsp; i++) splits[i+split*numsp] = 0;
      splits[r+split*numsp] = 1;
    }
    else
      for(i = 0; i < numsp; i++) splits[i+split*numsp] = splits[i+(r-numsp)*numsp];
    if (l < numsp)
      splits[l+split*numsp] = 1;
    else
      for(i = 0; i < numsp; i++) splits[i+split*numsp] += splits[i+(l-numsp)*numsp];
  }
  for(split = 0; split < numsp-3; split++)
    split_leading1(numsp,&splits[split*numsp]);
}
double dim_split(int numsp, int ntrees){
  /* integer 2^{numsp-1}-numsp-1 but too large numsp will convert to negative 
   * number */
  double max1=0;
/*   int i,max2=1; */
  max1 = 5*ntrees*(numsp-1);
/*   for(i = 1; i <= numsp-1; i++) max2 *= 2; */
/*   max2 -= (numsp + 1); */
/*   max1=(max1 < max2)?max1:max2; */
  return(max1);
}
int ivsame(int n, int *x, int *y){
  int i;
  for(i = 0; i < n; i++)
    if (x[i] != y[i]) return(0);
  return(1);
}
void backward_prob(int nchar, int ntaxa, int l, int *s, double *M, double *f, 
		   double *b){
  int i,j;
  double bt;
  if (l < ntaxa && s[l] < nchar)
    for(i = 0; i < nchar; i++) b[i] *= M[i+s[l]*nchar];
  if (l >= ntaxa)
    for(i = 0; i < nchar; i++){
      bt = 0.0;
      for(j = 0; j < nchar; j++) bt += M[i+j*nchar]*f[j];
      b[i] *= bt;
    }
/*   for(i = 0; i < nchar; i++) printf("%.3e ", b[i]); */
/*   printf("\n"); */
}
void ocalc(int nchar, double *M, double *b, double *o){
  int i,j;
  for(i = 0; i < nchar; i++){
    o[i] = 0.0;
    for(j = 0; j < nchar; j++)
      o[i] += M[i+j*nchar]*b[j];
  }
}
void backward_probs(int nchar, int ntaxa, int *s, double *utreec, double *M, 
		    double *f, double *b, int deriv){
  /* b[i+r*nchar] gives prob, starting from i, of data "backward" from r, not
   * including t_r */ 
  int nchar2,i,k,r,l;
  nchar2 = nchar*nchar;
  r = utreec[(ntaxa-2)*4]; l = utreec[1+(ntaxa-2)*4]; /* root */
  for(i = 0; i < nchar; i++) b[i+r*nchar] = f[i+(l-ntaxa)*nchar];
  for(i = 0; i < nchar; i++) b[i+l*nchar] = 1;
  backward_prob(nchar,ntaxa,r,s,&M[r*nchar2],&f[(r-ntaxa)*nchar],&b[l*nchar]);
/*   printf("\n"); */
/*   for(i = 0; i < nchar; i++){ */
/*     for(k = 0; k < nchar; k++) printf("%5e ", M[i+k*nchar+r*nchar2]); */
/*     printf("\n"); */
/*   } */
  if (!deriv) return;

  for(k = ntaxa-3; k >= 0; k--){
    r = (int) utreec[k*4]; l = (int) utreec[1+k*4];
    if (k == ntaxa-3) 
      for(i = 0; i < nchar; i++) b[i+r*nchar] = b[i+(k+ntaxa)*nchar];
    else
      ocalc(nchar,&M[(k+ntaxa)*nchar2],&b[(k+ntaxa)*nchar],&b[r*nchar]);
    for(i = 0; i < nchar; i++) b[i+l*nchar] = b[i+r*nchar];
    backward_prob(nchar,ntaxa,l,s,&M[l*nchar2],&f[(l-ntaxa)*nchar],&b[r*nchar]);
    backward_prob(nchar,ntaxa,r,s,&M[r*nchar2],&f[(r-ntaxa)*nchar],&b[l*nchar]);
  }
}
void forward_prob(int nchar, int ntaxa, int k, int r, int *s, double *M, 
		  double *fr, double *f){
  int i,j;
  double ft;
/*   printf("%i\n", r); */
/*   for(i = 0; i < nchar; i++){ */
/*     for(j = 0; j < nchar; j++) printf("%.3f ", M[i+j*nchar]); */
/*     printf("\n"); */
/*   } */
  if (r < ntaxa && s[r] < nchar) 
    for(i = 0; i < nchar; i++)
      f[i] *= M[i+s[r]*nchar];
  if (r >= ntaxa){
    for(i = 0; i < nchar; i++){
      ft = 0.0;
      for(j = 0; j < nchar; j++) ft += M[i+j*nchar]*fr[j];
      f[i] *= ft;
    }
  }
/*   printf("\n");for(i = 0; i < nchar; i++) printf("%.3e ", f[i]); */
/*   printf("\n"); */
}
void forward_probs(int nchar, int ntaxa, int *s, double *utreec, double *M, 
		   double *f){
  int nchar2,i,k,r,l;
  nchar2 = nchar*nchar;
/*   for(i = 0; i < ntaxa-1; i++) printf("%i %i %.3f %.3f\n", (int) utreec[i*4], (int) utreec[1+i*4], utreec[2+i*4], utreec[3+i*4]); */
  for(i = 0; i < nchar*(ntaxa-2); i++) f[i] = 1.0;
  for(k = 0; k < ntaxa-2; k++){
    r = (int) utreec[k*4]; l = (int) utreec[1+k*4];
    forward_prob(nchar,ntaxa,k-ntaxa,r,s,&M[r*nchar2],&f[(r-ntaxa)*nchar],
		 &f[k*nchar]);
    forward_prob(nchar,ntaxa,k-ntaxa,l,s,&M[l*nchar2],&f[(l-ntaxa)*nchar],
		 &f[k*nchar]);
  }
}
 void lik_deriv_calc(int nchar, int ntaxa, int *s, double *fr, double *utreec, 
		     double *Mp, double *f, double *b, double *lp){
   int nchar2,k,i,j;
   nchar2 = nchar*nchar;
   for(k = 0; k < ntaxa; k++){
     lp[k] = 0.0;
     if (s[k] < nchar)
       for(i = 0; i < nchar; i++)
	 lp[k] += b[i+k*nchar]*fr[i]*Mp[i+s[k]*nchar+k*nchar2];
   }
   for(k = ntaxa; k < 2*ntaxa-3; k++){
     lp[k] = 0.0;
     for(i = 0; i < nchar; i++)
       for(j = 0; j < nchar; j++)
	 lp[k] += b[i+k*nchar]*fr[i]*Mp[i+j*nchar+k*nchar2]*f[j+(k-ntaxa)*nchar];
   }
     
 }
double lik_deriv_site_1rate(int nchar, int ntaxa, int *s, double *fr, 
			    double *utreec, double *M, double *Mp, double *lp,
			    int deriv){
#define NCHAR 20
#define NTAXA 100
  int i;
  double lik;
  double *f,*b;
  f = (double *) malloc((size_t) nchar*ntaxa*sizeof(double));
  b = (double *) malloc((size_t) nchar*2*ntaxa*sizeof(double));
  forward_probs(nchar,ntaxa,s,utreec,M,f);
  backward_probs(nchar,ntaxa,s,utreec,M,f,b,deriv);
  lik = 0.0;
  for(i = 0; i < nchar; i++){
/*     printf("%e\n",fr[i]*f[i+(ntaxa-3)*nchar]*b[i+(2*ntaxa-3)*nchar]); */
    lik += fr[i]*f[i+(ntaxa-3)*nchar]*b[i+(2*ntaxa-3)*nchar];
  }
  if (deriv)
    lik_deriv_calc(nchar,ntaxa,s,fr,utreec,Mp,f,b,lp);
  free(f); free(b);
  return(lik);
}
double lik_deriv_site(int nchar, int ntaxa, int *s, double *fr, double *utreec, 
		      int nrate, double *rate, double *wt, 
		      double *M, double *Mp, double *lp, int deriv){
  int k,ir,nchar2,nb,derivl;
  double lik,*lpt,te;
  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  lik = 0.0;
  derivl=0;
  if (deriv>0){
    derivl=1;
    lpt = (double *) malloc((size_t) nb*sizeof(double));
    for(k = 0; k < nb; k++) lp[k] = 0.0;
    if(deriv==2) lp[nb]=0;
  }
  for(ir = 0; ir < nrate; ir++){
    lik += wt[ir]*lik_deriv_site_1rate(nchar,ntaxa,s,fr,utreec,&M[ir*nchar2*nb], 
				       &Mp[ir*nchar2*nb],lpt,derivl);
    if (deriv>0){
      for(k = 0; k < nb; k++) lp[k] += wt[ir]*lpt[k];
    }
    if(deriv==2){
      for(te=0.0, k=0; k<ntaxa-2; k++)
	te += utreec[2+k*4]*lpt[((int) utreec[k*4])]
	  +utreec[3+k*4]*lpt[((int) utreec[1+k*4])];
      te += utreec[2+(ntaxa-2)*4]*lpt[((int) utreec[(ntaxa-2)*4])];
      lp[nb] += wt[ir]*rate[ir+nrate]*te/rate[ir];
    }
  }
  if (deriv>0) free(lpt);
  return(lik);
}
double ederiv(int nchar, int ntaxa, int *s, double *fr, double *utreec,  
	    double *M, double *Mp, int e)
{
  /* IN:
   * s - site pattern
   * utreec: the tree, nutreec format
   * fr - frequencies of character states
   * M, Mp - nchar x nchar x ntaxa x nbranch giving transition probs 
   *          and derivs for all combinations of taxa and edges.
             For RAS, Mp should be rate * Mp for tree with edges x rate.
   * e - label of edge of interest
   */
  int k,l,r,nchar2,lup;
  double *f,d;

  f = (double *) malloc((size_t) nchar*(ntaxa-1)*sizeof(double));
  nchar2 = nchar*nchar;
  for(k = 0; k < nchar*(ntaxa-1); k++) f[k] = 1.0;
  for(k = 0; k < ntaxa-1; k++){
    lup = (k < ntaxa-2)?2:1;
    for(l = 0; l < lup; l++){
      r = (int) utreec[l+k*4]; 
      if (e != r)
	forward_prob(nchar,ntaxa,k-ntaxa,r,s,&M[r*nchar2],
		     &f[(r-ntaxa)*nchar],&f[k*nchar]);
      if (e == r)
	forward_prob(nchar,ntaxa,k-ntaxa,r,s,&Mp[r*nchar2],
		     &f[(r-ntaxa)*nchar],&f[k*nchar]);
    }
  }
  
  d = 0.0;
  for(k = 0; k < nchar; k++) 
    d += fr[k]*f[k+(ntaxa-2)*nchar]*f[k+(ntaxa-3)*nchar];
  free(f); 
  return(d);
}
double emderiv(int nchar, int ntaxa, int *s, double *fr, double *utreec, 
	       double *M, double *Mp, double *d, double *dm)
{
  /* IN:
   * s - site pattern
   * fr - frequencies of character states
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * M, Mp - nchar x nchar x nbranch giving transition probs and derivs 
   *          for all combinations of taxa and edges
   *         For RAS, Mp should be rate * Mp for tree with edges x rate.
   * OUT:
   * d - 2*ntaxa-4, derivatives for the external edges
   * dm - 3 dim, derivatives for the three topologies
   * RETURN: pattern probability contribution
   * COMMENTS:
   * original tree u[0]u[1]|u[2]u[3], where u[j] labels of edges at root split
   * top 0: orig, top 1: 02|13, top 2: 03|12
   */
  int nchar2,e,x,u[4],j,y,o[2];
  double *f,px,*su,*drt;

  f = (double *) malloc((size_t) nchar*ntaxa*sizeof(double));
  su = (double *) malloc((size_t) nchar*4*sizeof(double));
  drt = (double *) malloc((size_t) nchar*sizeof(double));

  nchar2 = nchar*nchar;
  for(e = 0; e < 2*ntaxa-4; e++) /* external derivatives */
    d[e] = ederiv(nchar,ntaxa,s,fr,utreec,M,Mp,e);
  forward_probs(nchar,ntaxa,s,utreec,M,f);
  px = 0.0; /* pattern probs */
  for(x = 0; x < nchar; x++) 
    px += fr[x]*f[x+(ntaxa-4)*nchar]*f[x+(ntaxa-3)*nchar];

  /* original tree is u[0]u[1]|u[2]u[3] */
  u[0] = (int) utreec[(ntaxa-4)*4]; u[1] = (int) utreec[1+(ntaxa-4)*4];
  u[2] = (int) utreec[(ntaxa-3)*4]; u[3] = (int) utreec[1+(ntaxa-3)*4];

  /* su give the cond probs of data for the u edges
   * similar to forward probs but includes the edge */
  for(j = 0; j < 4; j++)
    for(x = 0; x < nchar; x++){
      su[x+j*nchar] = 0.0;
      if (u[j] >= ntaxa)
	for(y = 0; y < nchar; y++) 
	  su[x+j*nchar] += M[x+y*nchar+u[j]*nchar2]*f[y+(u[j]-ntaxa)*nchar];
      if (u[j] < ntaxa)
	su[x+j*nchar] += M[x+s[u[j]]*nchar+u[j]*nchar2];
    }

  for(j = 0; j < 3; j++){ /* terminal derivatives */
    /* tree is u[0]u[j+1] | others; others in o */
    o[0] = 2; o[1] = 3;
    if (j == 1) o[0] = 1;
    if (j == 2) o[1] = 1;

    for(x = 0; x < nchar; x++){
      drt[x] = 0.0;
      for(y = 0; y < nchar; y++)
	drt[x] += Mp[x+y*nchar+(2*ntaxa-4)*nchar2]*su[y]*su[y+(j+1)*nchar];
    }

    dm[j] = 0.0;
    for(x = 0; x < nchar; x++)
      dm[j] += fr[x]*su[x+o[0]*nchar]*su[x+o[1]*nchar]*drt[x];
  }

  free(f); free(su); free(drt); 
  return(px);
}
double infomx3(int nchar, int ntaxa, int *s, double *fr, double *utreec, 
	       int nrate, double *wt, double *M, double *Mp, double *l2)
{
  /* IN:
   * s - site pattern
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * fr - frequencies of character states
   * M, Mp - nchar x nchar x nbranch x nrate giving transition probs and derivs 
              for all combinations of taxa and edges
   *         For RAS, Mp should be rate * Mp for tree with edges x rate
   * OUT:
   * l2 - numerator of information matrix for site pattern 
   *      (igorning second deriv part)
   *      (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   * RETURN: pattern probability
   * COMMENTS:
   * original tree u[0]u[1]|u[2]u[3], where u[j] labels of edges at root split
   * top 0: orig, top 1: 02|13, top 2: 03|12; see emderiv
   */
  int l,nchar2,j,k,nb;
  double *d,*dt,px,dm[3],dmt[3];
  
  d = (double *) malloc((size_t) 2*ntaxa*sizeof(double));
  dt = (double *) malloc((size_t) 2*ntaxa*sizeof(double));

  for(j = 0; j < 2*ntaxa-4; j++) d[j]=0.0;
  for(j = 0; j < 3; j++) dm[j] = 0.0;
  px = 0.0;
  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  for(l = 0; l < nrate; l++){
    px += wt[l]*emderiv(nchar,ntaxa,s,fr,utreec,&M[l*nchar2*nb],
			&Mp[l*nchar2*nb],dt,dmt);
    for(j = 0; j < 2*ntaxa-4; j++) d[j] += wt[l]*dt[j];
    for(j = 0; j < 3; j++) dm[j] += wt[l]*dmt[j];
  }
  /* for(j=0; j<4; j++) printf(" % .4e",d[j]); */
  /* for(j=0; j<3; j++) printf(" % .4e",dm[j]); printf("\n"); */

  for(j = 0; j < 2*ntaxa-4; j++)
    for(k = j; k < 2*ntaxa-4; k++){
      l2[j+k*(2*ntaxa-1)] =  d[j]*d[k];
      l2[k+j*(2*ntaxa-1)] = l2[j+k*(2*ntaxa-1)];
    }
  for(j = 0; j < 2*ntaxa-4; j++)
    for(k = 2*ntaxa-4; k < 2*ntaxa-1; k++){
      l2[j+k*(2*ntaxa-1)] = d[j]*dm[k-(2*ntaxa-4)];
      l2[k+j*(2*ntaxa-1)] = l2[j+k*(2*ntaxa-1)];
    }
  for(j = 2*ntaxa-4; j < 2*ntaxa-1; j++)
    for(k = 2*ntaxa-4; k < 2*ntaxa-1; k++){
      l2[j+k*(2*ntaxa-1)] = dm[j-(2*ntaxa-4)]*dm[k-(2*ntaxa-4)];
      l2[k+j*(2*ntaxa-1)] = l2[j+k*(2*ntaxa-1)];
    }
  free(d); free(dt);
  return(px);
}
double escorex3(int nchar, int ntaxa, int *s, double *fr, double *utreec, 
	       int nrate, double *wt, double *M, double *Mp, double *l1)
{
  /* IN:
   * s - site pattern
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * fr - frequencies of character states
   * M, Mp - nchar x nchar x nbranch x nrate giving transition probs and derivs 
              for all combinations of taxa and edges
   *         For RAS, Mp should be rate * Mp for tree with edges x rate
   * OUT:
   * l1 -  (2 ntaxa - 1) numerator of score for site pattern
   * RETURN: pattern probability
   * COMMENTS:
   * original tree u[0]u[1]|u[2]u[3], where u[j] labels of edges at root split
   * top 0: orig, top 1: 02|13, top 2: 03|12; see emderiv
   */
  int l,nchar2,j,nb;
  double *dt,px,dmt[3];
  
  dt = (double *) malloc((size_t) 2*ntaxa*sizeof(double));
  px = 0.0;
  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  for(j = 0; j < 2*ntaxa-1; j++) l1[j] = 0.0;
  for(l = 0; l < nrate; l++){
    px += wt[l]*emderiv(nchar,ntaxa,s,fr,utreec,&M[l*nchar2*nb],
			&Mp[l*nchar2*nb],dt,dmt);
    for(j = 0; j < 2*ntaxa-4; j++) l1[j] += wt[l]*dt[j];
    for(j = 0; j < 3; j++) l1[j+2*ntaxa-4] += wt[l]*dmt[j];
  }

  free(dt);
  return(px);
}
void infomatrixf3(int nchar, int ntaxa,double *fr, double *utreec, int nrate,
		  double *wt, double *M, double *Mp, double *Mm, 
		  double *infom, int mid)
{
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * fr - frequencies of character states
   * M, Mp - nchar x nchar x nbranch x nrate giving transition probs 
   *          and derivs for all combinations of taxa and edges 
   *         For RAS, Mp should be rate * Mp for tree with edges x rate
   * Mm - nchar x nchar x rate giving transition probs  for middle edge; not
   *      used if mid = 0
   * mid = 1 => Use middle branch in probability weighting
   * OUT:
   * infom:  information matrix for site pattern (igorning second deriv part)
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   */
  int npattf,i,*patt,x,k,j,*s,nchar2,nb;
  double px,*infomc,*Mmm,pxt,one=1;

  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  s = (int *) malloc((size_t) ntaxa*sizeof(int));
  infomc = (double *) malloc((size_t) (2*ntaxa-1)*(2*ntaxa-1)*sizeof(double));
  if (mid){
    Mmm = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
    memcpy(Mmm,M,nchar2*nb*nrate*sizeof(double));
    for(j = 0; j < nrate; j++)
      for(i = 0; i < nchar2; i++) 
	Mmm[i+(2*ntaxa-4)*nchar2+j*nchar2*nb]=Mm[i+j*nchar2];
  }
  npattf = 1;
  for(i = 0; i < ntaxa; i++) npattf *= nchar;
  patt = (int *) malloc((size_t) ntaxa*npattf*sizeof(int));
  npattf = all_sitepatt(nchar,ntaxa,patt);

  for(i = 0; i < (2*ntaxa-1)*(2*ntaxa-1); i++) infom[i] = 0.0;
  for(x = 0; x < npattf; x++){
    for(k = 0; k < ntaxa; k++) s[k] = patt[x+k*npattf];
    px = infomx3(nchar,ntaxa,s,fr,utreec,nrate,wt,M,Mp,infomc);
    if (mid){
      pxt = lik_deriv_site(nchar,ntaxa,s,fr,utreec,nrate,&one,wt,Mmm,Mp,Mp,0);
    }
    else{
      pxt = px;
    }
    if(pxt > 0.0)
      for(i = 0; i < (2*ntaxa-1); i++)
	for(j = i; j < (2*ntaxa-1); j++)
	  infom[i+j*(2*ntaxa-1)] += pxt*infomc[i+j*(2*ntaxa-1)]/(px*px);
  }
  for(i = 0; i < (2*ntaxa-1); i++)
      for(j = i; j < (2*ntaxa-1); j++)
	infom[j+i*(2*ntaxa-1)] = infom[i+j*(2*ntaxa-1)];
  free(patt); free(s); free(infomc);
  if (mid) free(Mmm);
}
void infomatrixf3_patt(int nchar, int ntaxa,double *fr, double *utreec, 
		       int nrate, double *wt,
		       double *M, double *Mp, double *infom, 
		       int nsite, int npattf, int *patt, int *count)
{
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * fr - frequencies of character states
   * M, Mp - nchar x nchar x nbranch x nrate giving transition probs 
   *          and derivs for all combinations of taxa and edges 
   *         For RAS, Mp should be rate * Mp for tree with edges x rate
   * patt - npattf x ntaxa giving corresp patterns for count
   * nsite sites used in obtaining count
   * OUT:
   * infom:  information matrix for site pattern (igorning second deriv part)
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   */
  int i,x,k,j,*s;
  double px,*infomc,pxt;

  s = (int *) malloc((size_t) ntaxa*sizeof(int));
  infomc = (double *) malloc((size_t) (2*ntaxa-1)*(2*ntaxa-1)*sizeof(double));

  /* printf("%i %i %i %i %i\n", nchar, ntaxa,nrate,nsite,npattf);  */
  for(i = 0; i < (2*ntaxa-1)*(2*ntaxa-1); i++) infom[i] = 0.0;
  for(x = 0; x < npattf; x++){
    for(k = 0; k < ntaxa; k++) s[k] = patt[x+k*npattf];
    /* printf("%5i ",x); for(i=0; i<ntaxa; i++) printf("%i",s[i]); */
    px = infomx3(nchar,ntaxa,s,fr,utreec,nrate,wt,M,Mp,infomc);
    pxt = count[x] / ((double) nsite);
    for(i = 0; i < (2*ntaxa-1); i++)
      for(j = i; j < (2*ntaxa-1); j++)
	infom[i+j*(2*ntaxa-1)] += pxt*infomc[i+j*(2*ntaxa-1)]/(px*px);
  }

  for(i = 0; i < (2*ntaxa-1); i++)
      for(j = i; j < (2*ntaxa-1); j++)
	infom[j+i*(2*ntaxa-1)] = infom[i+j*(2*ntaxa-1)];
  free(s); free(infomc);
}
void escoref3(int nchar, int ntaxa,double *fr, double *utreec, int nrate,
		  double *wt, double *M, double *Mp, double *Mm, 
		  double *escore, int mid)
{
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * fr - frequencies of character states
   * M, Mp - nchar x nchar x nbranch x nrate giving transition probs 
   *          and derivs for all combinations of taxa and edges 
   *         For RAS, Mp should be rate * Mp for tree with edges x rate
   * Mm - nchar x nchar x rate giving transition probs  for middle edge; not
   *      used if mid = 0
   * mid = 1 => Use middle branch in probability weighting
   * OUT:
   * escore: (2 ntaxa - 1) first (2 ntaxa - 4) external then middle
   */
  int npattf,i,j,*patt,x,k,*s,nchar2,nb;
  double px,*escorec,*Mmm,pxt,one=1;
  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  s = (int *) malloc((size_t) ntaxa*sizeof(int));
  escorec = (double *) malloc((size_t) (2*ntaxa-1)*sizeof(double));
  if (mid){
    Mmm = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
    memcpy(Mmm,M,nchar2*nb*nrate*sizeof(double));
    for(j = 0; j < nrate; j++)
      for(i = 0; i < nchar2; i++) 
	Mmm[i+(2*ntaxa-4)*nchar2+j*nchar2*nb]=Mm[i+j*nchar2];
  }
  npattf = 1;
  for(i = 0; i < ntaxa; i++) npattf *= nchar;
  patt = (int *) malloc((size_t) ntaxa*npattf*sizeof(int));
  npattf = all_sitepatt(nchar,ntaxa,patt);

  for(i = 0; i < (2*ntaxa-1); i++) escore[i] = 0.0;
  for(x = 0; x < npattf; x++){
    for(k = 0; k < ntaxa; k++) s[k] = patt[x+k*npattf];
    px = escorex3(nchar,ntaxa,s,fr,utreec,nrate,wt,M,Mp,escorec);
    if (mid){
      pxt = lik_deriv_site(nchar,ntaxa,s,fr,utreec,nrate,&one,wt,Mmm,Mp,Mp,0);
      /* printf("px %e pxt %e\n",px,pxt); */
      /* wmat(stdout,nchar,nchar,&M[(2*ntaxa-4)*nchar2]); */
      /* printf("\n"); */
      /* wmat(stdout,nchar,nchar,&Mmm[(2*ntaxa-4)*nchar2]); */
      /* exit(0); */
      for(j = 0; j < nrate; j++)
	for(i = 0; i < nchar2; i++) 
	  Mmm[i+(2*ntaxa-4)*nchar2+j*nchar2*nb]=Mm[i+j*nchar2];
    }
    else{
      pxt = px;
    }
    for(i = 0; i < (2*ntaxa-1); i++)
      escore[i] += pxt*escorec[i]/px;
  }

  free(patt); free(s); free(escorec);
  if (mid) free(Mmm);
}
void escoref3_patt(int nchar, int ntaxa,double *fr, double *utreec, int nrate,
		   double *wt, double *M, double *Mp, 
		   double *escore,  int nsite, int npattf, int *patt, 
		   int *count)
{
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * fr - frequencies of character states
   * M, Mp - nchar x nchar x nbranch x nrate giving transition probs 
   *          and derivs for all combinations of taxa and edges 
   *         For RAS, Mp should be rate * Mp for tree with edges x rate
   * patt - npattf x ntaxa giving corresp patterns for count
   * nsite sites used in obtaining count
   * OUT:
   * escore: (2 ntaxa - 1) first (2 ntaxa - 4) external then middle
   */
  int i,x,k,*s;
  double px,*escorec,pxt;
  s = (int *) malloc((size_t) ntaxa*sizeof(int));
  escorec = (double *) malloc((size_t) (2*ntaxa-1)*sizeof(double));
  for(i = 0; i < (2*ntaxa-1); i++) escore[i] = 0.0;
  for(x = 0; x < npattf; x++){
    for(k = 0; k < ntaxa; k++) s[k] = patt[x+k*npattf];
    px = escorex3(nchar,ntaxa,s,fr,utreec,nrate,wt,M,Mp,escorec);
    pxt = count[x] / ((double) nsite);
    for(i = 0; i < (2*ntaxa-1); i++)
      escore[i] += pxt*escorec[i]/px;
  }

  free(s); free(escorec);
}
void infomatrixscore_gen(int nchar, int ntaxa, double *utreec, int nrate, 
			 double *rate, double *wt, 		      
			 double pij(int i, int j, double t, int nchar, 
				    double *fr, double *W, double *Lam),
			 double ppij(int i, int j, double t, int nchar, 
				     double *fr, double *W, double *Lam),
			 double *fr, double *W, double *Lam, double *infom, 
			 double *escore, int mid)
{
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * nrate, rate, wt: rate distribution
   * pij, ppij - substition prob function and derivatives,
   * fr, W, Lam  - dependent on substitution model
   * mid = 1 => Use middle branch in probability weighting
   * OUT:
   * infom:  information matrix for site pattern (igorning second deriv part)
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   * escore: (2 ntaxa - 1) first (2 ntaxa - 4) external then middle
   */
  int nchar2,j,k,l,nb,x,y;
  double *M,*Mp,*Mm,*utreecn,tm;

  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  M = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
  Mp = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
  Mm = (double *) malloc((size_t) nchar2*nrate*sizeof(double));

  tm = utreec[(ntaxa-2)*4+2];
  utreecn = (double *) malloc((size_t) (ntaxa-1)*4*sizeof(double));

  for(l = 0; l < (ntaxa-1)*4; l++) utreecn[l] = utreec[l];
  for(j = 0; j < nrate; j++){
    for(k = 0; k < ntaxa-2; k++)
      for(l = 0; l < 2; l++) utreecn[l+2+4*k] = utreec[l+2+4*k]*rate[j];
    for(l = 0; l < 2; l++) utreecn[l+2+4*(ntaxa-2)]=0.0;
    trans_gen(nchar,ntaxa,utreecn,fr,W,Lam,pij,ppij,&M[j*nchar2*nb],
		&Mp[j*nchar2*nb]);
    for(l = 0; l < nchar2*nb; l++) Mp[l+j*nchar2*nb] *= rate[j];
    for(x = 0; x < nchar; x++)
      for(y = 0; y < nchar; y++)
    	Mm[x+y*nchar+j*nchar2]=pij(x,y,tm*rate[j],nchar,fr,W,Lam);
  }
  infomatrixf3(nchar,ntaxa,fr,utreec,nrate,wt,M,Mp,Mm,infom,mid);
  escoref3(nchar,ntaxa,fr,utreec,nrate,wt,M,Mp,Mm,escore,mid);
  free(M); free(Mp); free(Mm); free(utreecn);
}
void infomatrix_gen_patt(int nchar, int ntaxa, double *utreec,  
			 int nrate, double *rate, double *wt, 		      
			 double pij(int i, int j, double t, int nchar, 
				    double *fr, double *W, double *Lam),
			 double ppij(int i, int j, double t, int nchar, 
				     double *fr, double *W, double *Lam),
			 double *fr, double *W, double *Lam, double *infom, 
			 int nsite, int npatt, int *patt, int *count)
{
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * nrate, rate, wt: rate distribution
   * pij, ppij - substition prob function and derivatives,
   * fr, W, Lam  - dependent on substitution model
   * patt - npatt x ntaxa giving corresp patterns for count
   * nsite sites used in obtaining count
   * OUT:
   * infom:  information matrix for site pattern (igorning second deriv part)
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   */
  int nchar2,j,k,l,nb;
  double *M,*Mp,*utreecn;

  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  M = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
  Mp = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
  utreecn = (double *) malloc((size_t) (ntaxa-1)*4*sizeof(double));

  for(l = 0; l < (ntaxa-1)*4; l++) utreecn[l] = utreec[l];
  for(j = 0; j < nrate; j++){
    for(k = 0; k < ntaxa-2; k++)
      for(l = 0; l < 2; l++) utreecn[l+2+4*k] = utreec[l+2+4*k]*rate[j];
    for(l = 0; l < 2; l++) utreecn[l+2+4*(ntaxa-2)]=0.0;
    trans_gen(nchar,ntaxa,utreecn,fr,W,Lam,pij,ppij,&M[j*nchar2*nb],
		&Mp[j*nchar2*nb]);
    for(l = 0; l < nchar2*nb; l++) Mp[l+j*nchar2*nb] *= rate[j];
  }
  infomatrixf3_patt(nchar,ntaxa,fr,utreec,nrate,wt,M,Mp,infom,
		    nsite,npatt,patt,count);
  free(M); free(Mp); free(utreecn);
}
void infomatrixscore_gen_patt(int nchar, int ntaxa, double *utreec,  
			      int nrate, double *rate, double *wt, 
			      double pij(int i, int j, double t, int nchar, 
					 double *fr, double *W, double *Lam),
			      double ppij(int i, int j, double t, int nchar, 
					  double *fr, double *W, double *Lam),
			      double *fr, double *W, double *Lam, double *infom,
			      double *escore,
			      int nsite, int npatt, int *patt, int *count)
{
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * nrate, rate, wt: rate distribution
   * pij, ppij - substition prob function and derivatives,
   * fr, W, Lam  - dependent on substitution model
   * patt - npatt x ntaxa giving corresp patterns for count
   * nsite sites used in obtaining count
   * OUT:
   * infom:  information matrix for site pattern (igorning second deriv part)
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   * escore: (2 ntaxa - 1) first (2 ntaxa - 4) external then middle
   */
  int nchar2,j,k,l,nb;
  double *M,*Mp,*utreecn;

  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  M = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
  Mp = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
  utreecn = (double *) malloc((size_t) (ntaxa-1)*4*sizeof(double));

  for(l = 0; l < (ntaxa-1)*4; l++) utreecn[l] = utreec[l];
  for(j = 0; j < nrate; j++){
    for(k = 0; k < ntaxa-2; k++)
      for(l = 0; l < 2; l++) utreecn[l+2+4*k] = utreec[l+2+4*k]*rate[j];
    for(l = 0; l < 2; l++) utreecn[l+2+4*(ntaxa-2)]=0.0;
    trans_gen(nchar,ntaxa,utreecn,fr,W,Lam,pij,ppij,&M[j*nchar2*nb],
		&Mp[j*nchar2*nb]);
    for(l = 0; l < nchar2*nb; l++) Mp[l+j*nchar2*nb] *= rate[j];
  }

  infomatrixf3_patt(nchar,ntaxa,fr,utreec,nrate,wt,M,Mp,infom,
		    nsite,npatt,patt,count);
  escoref3_patt(nchar,ntaxa,fr,utreec,nrate,wt,M,Mp,escore,
		nsite,npatt,patt,count);
  free(M); free(Mp); free(utreecn);
}
void infomatrix_gen(int nchar, int ntaxa, double *utreec,  
		    int nrate, double *rate, double *wt, 		      
		    double pij(int i, int j, double t, int nchar, 
			       double *fr, double *W, double *Lam),
		    double ppij(int i, int j, double t, int nchar, 
				double *fr, double *W, double *Lam),
		    double *fr, double *W, double *Lam, double *infom)
{
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * nrate, rate, wt: rate distribution
   * pij, ppij - substition prob function and derivatives,
   * fr, W, Lam  - dependent on substitution model
   * OUT:
   * infom:  information matrix for site pattern (igorning second deriv part)
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   */
  int nchar2,j,k,l,nb;
  double *M,*Mp,*utreecn,Mm=1.0;

  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  M = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
  Mp = (double *) malloc((size_t) nchar2*nb*nrate*sizeof(double));
  utreecn = (double *) malloc((size_t) (ntaxa-1)*4*sizeof(double));

  for(l = 0; l < (ntaxa-1)*4; l++) utreecn[l] = utreec[l];
  for(j = 0; j < nrate; j++){
    for(k = 0; k < ntaxa-2; k++)
      for(l = 0; l < 2; l++) utreecn[l+2+4*k] = utreec[l+2+4*k]*rate[j];
    for(l = 0; l < 2; l++) utreecn[l+2+4*(ntaxa-2)]=0.0;
    trans_gen(nchar,ntaxa,utreecn,fr,W,Lam,pij,ppij,&M[j*nchar2*nb],
		&Mp[j*nchar2*nb]);
    for(l = 0; l < nchar2*nb; l++) Mp[l+j*nchar2*nb] *= rate[j];
  }
  infomatrixf3(nchar,ntaxa,fr,utreec,nrate,wt,M,Mp,&Mm,infom,0);
  free(M); free(Mp); free(utreecn);
}
void covscorecor(int ntaxa, double *infom, double *Sc)
{
  /* IN:
   * infom: information matrix for site pattern (igorning second deriv part)
   *        format as in infomatrixf3, infomatrixJC
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   * OUT:
   * Sc: 3 x 3 matrix giving covariances, corrected for estimation 
   */ 
  int i,j;
  double *Ie,*w,y;

  Ie = (double *) malloc((size_t) (2*ntaxa-4)*(2*ntaxa-4)*sizeof(double));
  w = (double *) malloc((size_t) (2*ntaxa-4)*sizeof(double));

  for(i = 0; i < 2*ntaxa-4; i++)
    for(j = 0; j < 2*ntaxa-4; j++) 
      Ie[i+j*(2*ntaxa-4)]=infom[i+j*(2*ntaxa-1)];
  if (chol(2*ntaxa-4,Ie)){
    printf("External edge information matrix appears singular\n");
    exit(1);
  }
  
  for(i = 0; i < 3; i++){
    memcpy(w,&infom[(2*ntaxa-4+i)*(2*ntaxa-1)],(2*ntaxa-4)*sizeof(double));
    cholf(2*ntaxa-4,Ie,w);
    for(j = i; j < 3; j++){
      y = dot(2*ntaxa-4,&infom[(2*ntaxa-4+j)*(2*ntaxa-1)],1,w,1);
      Sc[i+j*3] = infom[(i+2*ntaxa-4)+(j+2*ntaxa-4)*(2*ntaxa-1)] - y;
      Sc[j+i*3] = Sc[i+j*3];
    }}
  free(Ie); free(w);
}
void covscorecore(int ntaxa, double *infom, double *escore, double *Sc, 
		  double *Ec)
{
  /* IN:
   * infom: information matrix for site pattern (igorning second deriv part)
   *        format as in infomatrixf3, infomatrixJC
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   * escore: (2 ntaxa - 1) first (2 ntaxa - 4) external then middle
   * OUT:
   * Sc: 3 x 3 matrix giving covariances, corrected for estimation 
   * Ec: E[V_jn^c]/sqrt(nsite)
   */ 
  int i,j;
  double *Ie,*w,*e,y;

  Ie = (double *) malloc((size_t) (2*ntaxa-4)*(2*ntaxa-4)*sizeof(double));
  w = (double *) malloc((size_t) (2*ntaxa-4)*sizeof(double));
  e = (double *) malloc((size_t) (2*ntaxa-4)*sizeof(double));

  for(i = 0; i < 2*ntaxa-4; i++)
    for(j = 0; j < 2*ntaxa-4; j++) 
      Ie[i+j*(2*ntaxa-4)]=infom[i+j*(2*ntaxa-1)];
  if (chol(2*ntaxa-4,Ie)){
    printf("External edge information matrix appears singular\n");
    exit(1);
  }
  
  for(i = 0; i < 3; i++){
    memcpy(w,&infom[(2*ntaxa-4+i)*(2*ntaxa-1)],(2*ntaxa-4)*sizeof(double));
    cholf(2*ntaxa-4,Ie,w);
    memcpy(e,escore,(2*ntaxa-4)*sizeof(double));
    cholf(2*ntaxa-4,Ie,e);
    y = dot(2*ntaxa-4,&infom[(2*ntaxa-4+i)*(2*ntaxa-1)],1,e,1);
    Ec[i] = escore[(2*ntaxa-4)+i] - y;
    for(j = i; j < 3; j++){
      y = dot(2*ntaxa-4,&infom[(2*ntaxa-4+j)*(2*ntaxa-1)],1,w,1);
      Sc[i+j*3] = infom[(i+2*ntaxa-4)+(j+2*ntaxa-4)*(2*ntaxa-1)] - y;
      Sc[j+i*3] = Sc[i+j*3];
    }}
  for(i = 0; i < 3; i++) Ec[i] /= sqrt(Sc[i+i*3]);
  free(Ie); free(w); free(e);
}
void corscorecor(int ntaxa, double *infom, double *Sc, double *Ic)
{
  /* IN:
   * infom: information matrix for site pattern (igorning second deriv part)
   *        format as in infomatrixf3, infomatrixJC
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   * OUT:
   * Sc: 3 x 3 matrix giving correlations, corrected for estimation 
   * Ic: 3 dimensional giving corrected informations
   */
  int i,j;
  covscorecor(ntaxa,infom,Sc);
  for(j = 0; j < 3; j++) Ic[j] = Sc[j+j*3];
  for(i = 0; i < 3; i++) /* convert to correlation matrix */
    for(j = 0; j < 3; j++) Sc[i+j*3] /= sqrt(Ic[j]*Ic[i]);
}
void corscorecore(int ntaxa, double *infom, double *escore, double *Sc, 
		  double *Ec, double *Ic)
{
  /* IN:
   * infom: information matrix for site pattern (igorning second deriv part)
   *        format as in infomatrixf3, infomatrixJC
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   * escore: (2 ntaxa - 1) first (2 ntaxa - 4) external then middle
   * OUT:
   * Sc: 3 x 3 matrix giving correlations, corrected for estimation 
   * Ic: 3 dimensional giving corrected informations
   * Ec: sqrt(nsite)*E[V_jn^c]
   */
  int i,j;
  covscorecore(ntaxa,infom,escore,Sc,Ec);
  for(j = 0; j < 3; j++) Ic[j] = Sc[j+j*3];
  for(i = 0; i < 3; i++) /* convert to correlation matrix */
    for(j = 0; j < 3; j++) Sc[i+j*3] /= sqrt(Ic[j]*Ic[i]);
}
void deltaxa(int ntaxa, char *taxon, char names[][11], double *utreec){
  char tname[11];
  int i,j,k,r,join,neighb,p,l;
  double tr,tn,tl;
   
  /* pad name with blanks */
  strncpy(tname,taxon,11);
  for(i=0; i<10; i++)
    if(tname[i]=='\0') break;
  for(j=i; j<10; j++) tname[j]=' ';
  tname[10]='\0';

  /* obtain label i of taxa */
  for(i=0; i<ntaxa; i++){
    if(strcmp(tname,names[i])==0) break;
  }
  if(i==ntaxa){
    printf("%s is not among the names in the treefile",taxon);
    exit(1);
  }

  /* find taxa i in tree */
  for(join=0; join<ntaxa-1; join++){
    r=utreec[join*4]; tr=utreec[2+join*4];
    l=utreec[1+join*4]; tl=utreec[3+join*4];
    /* printf("%i %i %i %i\n",i,join,r,l); */
    if(r==i){
      neighb=l; tn=tl; p=join+ntaxa;
      break;
    }
    if(l==i){
      neighb=r; tn=tr; p=join+ntaxa;
      break;
    }
  }

  /* update tree */
  for(j=join+1; j<ntaxa-1; j++){
    for(k=0; k<2; k++){
      if(utreec[k+j*4]==p){ utreec[k+j*4]=neighb; utreec[2+k+j*4] += tn; }
      if(utreec[k+j*4]>p) utreec[k+j*4]--;
    }
    memcpy(&utreec[(j-1)*4],&utreec[j*4],4*sizeof(double));
  }
  utreec[2+(ntaxa-3)*4] += utreec[3+(ntaxa-3)*4];
  utreec[3+(ntaxa-3)*4] = 0;

  /* adjust taxa labels and names */
  for(join=0; join<ntaxa-2; join++)
    for(k=0; k<2; k++)
      if(utreec[k+join*4]>i) utreec[k+join*4]--;
  for(j=i+1; j<ntaxa; j++) memcpy(names[j-1],names[j],11*sizeof(char));
  return;
}
void pr4design_paramf(FILE *fp, FILE **utreecfile, FILE **treefile, 
		      FILE **frfile, FILE **Qfile, FILE **ratefile, 
		      FILE **seqfile, char *aaRmat, int *ntaxa, 
		      int *nchar, int *model, int *ncatG, int *nrate, 
		      int *tinfo, int *nsim, int *iseed, int *nsite, 
		      double *kappa, double *alpha, int *soi, double *tm,
		      int *condl, int *mid){
  char c,opt[100],value[100];
  int is_option,done=0;
  *utreecfile=NULL;
  *treefile=NULL;
  *frfile=NULL;
  *Qfile=NULL;
  *ratefile=NULL;
  *seqfile=NULL;
  *aaRmat='\0';
  *nchar=4;
  *nrate=1;
  *ncatG=4;
  *alpha=-1;
  *tinfo=0;
  *nsim=100000;
  *iseed=1;
  *condl=0;
  *mid=1;
  *tm=-1;

  *ntaxa=-1; /* required parameters */
  *model=-1;
  *soi=-1;
  *nsite=-1;
  while(!done){
    c = ctl_paraml(fp,opt,value,&is_option);
    if(is_option){
      /* printf("option: %s value: %s\n",opt, value); */
      if(strcmp(opt,"splitofinterest")==0) *soi=(int) atof(value);
      if(strcmp(opt,"middle")==0) *tm=atof(value);

      if(strcmp(opt,"ntaxa")==0) *ntaxa=atoi(value);
      if(strcmp(opt,"treefile")==0) 
	*treefile=fopene(value,"r","treefile not available");

      if(strcmp(opt,"nsite")==0) *nsite=(int) atof(value);

      if(strcmp(opt,"nchar")==0) *nchar=atoi(value);
      if(strcmp(opt,"model")==0) *model=atoi(value);
      if(strcmp(opt,"aaRatefile")==0) sprintf(aaRmat,"%s",value);
      if(strcmp(opt,"frfile")==0) 
	*frfile=fopene(value,"r","frfile not available");
      if(strcmp(opt,"Qfile")==0) 
	*Qfile=fopene(value,"r","Qfile not available");
      if(strcmp(opt,"kappa")==0) *kappa=atof(value);
      if(strcmp(opt,"ttratio")==0) *kappa=-atof(value);
      
      if(strcmp(opt,"alpha")==0) *alpha=atof(value);
      if(strcmp(opt,"ncatG")==0) *ncatG=atoi(value);
      if(strcmp(opt,"nrate")==0) *nrate=atoi(value);
      if(strcmp(opt,"ratefile")==0)
	*ratefile=fopene(value,"r","ratefile not available");

      if(strcmp(opt,"tinfo")==0) *tinfo=atoi(value);
      if(strcmp(opt,"seqfile")==0) 
	*seqfile=fopene(value,"r","seqfile not available");

      if(strcmp(opt,"nsim")==0) *nsim=(int) atof(value);
      if(strcmp(opt,"iseed")==0) *iseed=(int) atof(value);
      if(strcmp(opt,"conditional")==0) *condl=(int) atof(value);
      if(strcmp(opt,"mid")==0) *mid=(int) atof(value);
    }
    if(c==EOF) done=1;
  }
  if(*ntaxa < 0){printf("ntaxa must be indicated"); exit(1);}
  if(*model < 0){printf("model must be indicated"); exit(1);}
  if(*soi < 0){printf("splitofinterest must be indicated"); exit(1);}
  if(*nsite < 0){printf("nsite must be indicated"); exit(1);}
  if(*tinfo==2 && *seqfile==NULL){
    printf("sequence file must be indicated when tinfo = 2\n"); exit(1);
  }
  if(*tinfo>0) *mid=0;
}
void pr_short_name(FILE *fp, char *name){
  int lb = 9, i;
  while(name[lb] == ' ') lb--;
  for(i=0; i<lb; i++) putc(name[i],fp);
  putc(' ',fp);
}
void pr_name_split(FILE *fp, int ntaxa, char names[][11], int *split){
  int i;
  for(i=0; i<ntaxa; i++)
    if(split[i]==1) pr_short_name(fp,names[i]);
  fprintf(fp,"\nsplit from\n");
  for(i=0; i<ntaxa; i++)
    if(split[i]==0) pr_short_name(fp,names[i]);
  fprintf(fp,"\n");
}
double pr_corrtop(int ntaxa, int nsite, double tm, double *infom, 
		  double *escore, int mid, int condl)
{
  /* IN:
   * tm - middle edge
   * infom - information matrix for site pattern (igorning second deriv part)
   *         format as in infomatrixf3, infomatrixJC
   *        (2 ntaxa - 1) x (2 ntaxa - 1) 
   *      (2 ntaxa - 4) x (2 ntaxa - 4) 11 block gives info entries for 
   *        external edges
   *      (2 ntaxa - 4) x 3 12 (and 21^T) block gives info entries for 
   *        external edges and split of interest
   *      3 x 3 22 block gives info for split of interest
   * escore: (2 ntaxa - 1) first (2 ntaxa - 4) external then middle
   * mid = 1 => Use middle branch in probability weighting
   * condl - 1 implies probabilities are conditional upon a resolved estimated
   *         topology
   */
  int j,zero=0,i;
  double Sc[9],Ic[3],d[3],A[3],mu[3],prop,p,eps=1.0e-9;

  if (mid){
    corscorecore(ntaxa,infom,escore,Sc,mu,Ic); 
    for(j = 0; j < 3; j++) mu[j] *= sqrt(nsite);
  }
  else{
    corscorecor(ntaxa,infom,Sc,Ic);
    for(j = 0; j < 3; j++) mu[j] = Sc[j]*sqrt(nsite*Ic[0])*tm;
  }
  /* wrow(stdout,3,mu,1); */
  /* wmat(stdout,3,3,Sc); */
  
  d[0] = mu[0]-mu[1]; d[1] = mu[1]-mu[2]; d[2] = mu[0]; 
  d[0] /= sqrt(2.0*(1- Sc[0+1*3])); d[1] /= sqrt(2.0*(1-Sc[1+2*3]));
  A[0] = Sc[0+1*3]-Sc[0+2*3]-1+Sc[1+2*3];
  A[0] /= 2*sqrt((1- Sc[0+1*3])*(1- Sc[1+2*3]));
  A[1] = sqrt((1-Sc[0+1*3])/2);
  A[2] = Sc[0+1*3]-Sc[0+2*3];
  A[2] /= sqrt(2*(1-Sc[1+2*3]));
  tvtls_(&zero,d,A,&eps,&prop);
  
  d[0] = mu[0]-mu[2]; d[1] = mu[2]-mu[1]; d[2] = mu[0]; 
  d[0] /= sqrt(2.0*(1- Sc[0+2*3])); d[1] /= sqrt(2.0*(1-Sc[1+2*3]));
  A[0] = Sc[0+2*3]-Sc[0+1*3]-1+Sc[1+2*3];
  A[0] /= 2*sqrt((1- Sc[0+2*3])*(1- Sc[1+2*3]));
  A[1] = sqrt((1-Sc[0+2*3])/2);
  A[2] = Sc[0+2*3]-Sc[0+1*3];
  A[2] /= sqrt(2*(1-Sc[1+2*3]));
  tvtls_(&zero,d,A,&eps,&p);
  prop += p;

  if (condl){
    for(i = 0; i < 3; i++) d[i] = -mu[i];
    A[0] = Sc[0+1*3]; A[1] = Sc[0+2*3]; A[2] = Sc[1+2*3];
    tvtls_(&zero,d,A,&eps,&p);
    prop /= (1.0 - p);
  }

  return(prop);
}
double pr_corrtop_gen(int nchar, int ntaxa, int nsite, double *utreec,
		      int nrate, double *rate, double *wt, 
		      double pij(int i, int j, double t, int nchar, 
				 double *fr, double *W, double *Lam),
		      double ppij(int i, int j, double t, int nchar, 
				  double *fr, double *W, double *Lam),
		      double *fr, double *W, double *Lam, int condl, int mid,
		      int tinfo, int nsim, int *seq, int iseed, int smodel, 
		      double *Qk)
{ 
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * pij, ppij - substition prob function and derivatives,
   * fr, W, Lam  - dependent on substitution model
   * condl - 1:conditional probability, given well-resolved top is estimated
   * mid = 1 => Use middle branch in probability weighting
   *
   * Rate distribution: nrate, alpha
   * if rate[0] < 0, a discrete gamma model w/ param alpha=-rate[0] is used 
   * calculation of rates same as paml 4.2
   * If tinfo=1 below, a gamma model should be used
   *
   * Info calculation:
   * tinfo - 0:exact, 1:using seq-gen, 2:seq data in seq
   * nsim: number of sites to simulate or allocated, tinfo=1 or 2
   * seq - should contain sequence data if tinfo=2, 
   *       should be allocated if tinfo=1
   *       not used if tinfo=0
   * iseed: seed for seq-gen, tinfo=1
   * smodel, Qk: model and parameters for seq-gen, tinfo=1
   *
   * original tree u[0]u[1]|u[1]u[2], where u[j] labels of edges at root split
   * top 0: orig, top 1: 02|13, top 2: 03|12
   */
  double *infom,*escore,tm,pcorr,alpha;
  int npatt,*count,*seqt,midt,i;

  /* FILE *infomfile; */

  alpha=-rate[0];
  if (rate[0] < 0){
    DiscreteGamma(wt,rate,alpha,alpha,nrate,0);
    for(i = 0; i < nrate; i++) wt[i] = 1.0/nrate;
  }
  infom = (double *) malloc((size_t) (2*ntaxa-1)*(2*ntaxa-1)*sizeof(double));
  escore = (double *) malloc((size_t) (2*ntaxa-1)*sizeof(double));
  tm = utreec[2+(ntaxa-2)*4];
/*   pr_utreec(stdout,ntaxa,utreec); */
  /* using sequence data to approximate info */
  if (tinfo==1)
    seq_genc(smodel,ntaxa,utreec,nsim,seq,nrate,alpha,iseed,fr,Qk);

  midt=mid;
  if (tinfo==1 || tinfo==2){
    midt=1;
    seqt = (int *) malloc((size_t) nsim*ntaxa*sizeof(int));
    count = (int *) malloc((size_t) nsim*sizeof(int));
    memcpy(seqt,seq,nsim*ntaxa*sizeof(int));
    npatt=patt_freq(ntaxa,nsim,seqt,count);
    
    infomatrixscore_gen_patt(nchar,ntaxa,utreec,nrate,rate,wt,pij,ppij,fr,W,Lam,
			     infom,escore,nsim,npatt,seqt,count);
  }
  /* exact calculation */
  if(tinfo==0)
    infomatrixscore_gen(nchar,ntaxa,utreec,nrate,rate,wt,pij,ppij,fr,W,Lam,infom,
			escore,midt);
  /* wmat(stdout,(2*ntaxa-1),(2*ntaxa-1),infom); */
  /* infomfile=fopen("tmp.infom","wb"); */
  /* fwrite(infom,sizeof(double),(2*ntaxa-1)*(2*ntaxa-1),infomfile); */
  /* fclose(infomfile); */

  utreec[2+(ntaxa-2)*4]=tm;
  pcorr = pr_corrtop(ntaxa,nsite,tm,infom,escore,midt,condl);
  if(tinfo==1 || tinfo==2){free(seqt); free(count);}
  free(infom); free(escore); 
  return(pcorr);
}
double pr_corrtop_smodel(int nchar, int smodel, int ntaxa, int nsite, 
			 double *utreec, int nrate, double *rate, double *wt, 
			 double *fr, double *Qk, int condl, int mid,
			 int tinfo, int nsim, int *seq, int iseed){
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * nchar - number of character states
   * smodel - 0:JC, 1:F81, 2:F84, 3:GTR, 4:PAM, 5:JTT, 6:WAG, 7:mtREV24, 8:HKY,
   *          9:LG
   * Qk - rate matrix for GTR, K or Kappa for F84, HKY; not used otherwise
   * condl - 1:conditional probability, given well-resolved top is estimated
   *
   * Rate distribution: nrate, alpha
   * if rate[0] < 0, a discrete gamma model w/ param alpha=-rate[0] is used 
   * calculation of rates same as paml 4.2
   * If tinfo=1 below, a gamma model should be used
   *
   * Info calculation:
   * tinfo - 0:exact, 1:using seq-gen, 2:seq data in seq
   * nsim: number of sites to simulate or allocated, tinfo=1 or 2
   * seq - should contain sequence data if tinfo=2, 
   *       should be allocated if tinfo=1
   *       not used if tinfo=0
   * iseed: seed for seq-gen, tinfo=1
   *
   * IN/OUT:
   * fr - frequencies; reset for JC, determined from Q for GTR
   *      if *fr < 0 for aa models, model-based frequencies used 
   */
  int allocW=0;
  double *W,*Lam,pr;
  double (*pij)(int i, int j, double t, int nchar, double *fr, double *W, 
		double *Lam);
  double (*ppij)(int i, int j, double t, int nchar, double *fr, double *W,
		 double *Lam);
  allocW=subst_model_set(nchar,smodel,fr,Qk,&pij,&ppij,&W,&Lam);
  pr=pr_corrtop_gen(nchar,ntaxa,nsite,utreec,nrate,rate,wt,*pij,*ppij,fr,W,Lam,
		    condl,mid,tinfo,nsim,seq,iseed,smodel,Qk);
  if(allocW==1 || allocW==2) free(W);
  if(allocW==2) free(Lam);
  return(pr);
}
void pr_corrtop_mutreec(int nchar, int ntaxa, int nsite, int ntree,
			int nrate, double *rate, double *wt,
			double *utrees, double *p, int condl, int mid)
{ 
  /* IN/OUT:
   * utrees - 4 x (ntaxa-1) x ntree, giving the trees
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * condl - 1 implies probabilities are conditional upon a resolved estimated
   *         topology
   * mid = 1 => Use middle branch in probability weighting
   *
   * OUT:
   * p - ntree-dim vector of probabilities
   *
   * COMMENTS:
   * Does not allow approximation via simulation
   */
  int itree,not_used=-1;
  double tm,*fr;
  fr=(double *) malloc((size_t) nchar*sizeof(double));
  for(itree = 0; itree < ntree; itree++){
    tm = utrees[2+(ntaxa-2)*4+itree*4*(ntaxa-1)];
    /* utrees[2+(ntaxa-2)*4+itree*4*(ntaxa-1)]=0.0; */
    p[itree]=pr_corrtop_smodel(nchar,0,ntaxa,nsite,&utrees[itree*4*(ntaxa-1)],
			       nrate,rate,wt,fr,&tm,condl,mid,0,0,
			       &not_used,not_used);
    utrees[2+(ntaxa-2)*4+itree*4*(ntaxa-1)]=tm;
  }
}
void pr_corrtop_mn(int nchar, int ntaxa, int *nsite, double *utreec, 
		   int nrate, double *rate, double *wt,  
		   int nn, double *pn, int condl, int mid)
{ 
  /* IN:
   * utreec: the tree, nutreec format
   *    conventions: r < l, last edge-length 0
   *                 root is split of interest and has 
   *                     r l: (2 ntaxa - 4) (2 ntaxa - 3)
   * condl - 1 implies probabilities are conditional upon a resolved estimated
   *         topology
   * mid = 1 => Use middle branch in probability weighting
   * OUT:
   * pn - vector of probabilities; dimension nn
   */
  int in;
  double *infom,*escore,*fr,tm,W,Lam;

  infom = (double *) malloc((size_t) (2*ntaxa-1)*(2*ntaxa-1)*sizeof(double));
  escore = (double *) malloc((size_t) (2*ntaxa-1)*sizeof(double));
  fr = (double *) malloc((size_t) nchar*sizeof(double));

  tm = utreec[2+4*(ntaxa-2)];
  /* utreec[2+4*(ntaxa-2)] = 0.0; */
  for(in = 0; in < nchar; in++) fr[in]=1.0/nchar;
  infomatrixscore_gen(nchar,ntaxa,utreec,nrate,rate,wt,pijJCg_gen,ppijJCg_gen,fr,
		      &W,&Lam,infom,escore,mid);
  
  utreec[2+4*(ntaxa-2)] = tm;

  for(in = 0; in < nn; in++)
    pn[in]=pr_corrtop(ntaxa,nsite[in],tm,infom,escore,mid,condl);

  free(infom); free(escore); free(fr );
}
void makeprotfreqs(double *fr, double *W, double *Lam) 
{
  long i, mineig,nchar=20;
  mineig = 0;
  for (i = 0; i < nchar; i++)
    if (fabs(Lam[i]) < fabs(Lam[mineig]))
      mineig = i;
  for(i = 0; i < nchar; i++) fr[i] =  fabs(W[i+mineig*nchar]);
}
void Mijp(double t, double *fr, double *W, double *Lam, double *f){
  /* transition probs
   * adapted from proml make_pmatrix 
   */
  static double elambdat[NCHAR];
  double q, p0;
  int i, j, k;

  if (t == 0.0){
    for(i = 0; i < NCHAR; i++)
      for(j = 0; j < NCHAR; j++) f[i + j*NCHAR] = (i == j)?1.0:0.0;
    return;
  }

  for(k = 0; k < NCHAR; k++) elambdat[k] = exp(t * Lam[k]);
  for(i = 0; i < NCHAR; i++)
    for(j = 0; j < NCHAR; j++){      
      p0 = 0.0;
      for(k = 0; k < NCHAR; k++){
	q = W[i + k*NCHAR] * W[j+ k*NCHAR];
	p0 += (q * elambdat[k]);
      }
      f[i + j*NCHAR] = p0/fr[i];
    }

}
void Mijpd(double t, double *fr, double *W, double *Lam, double *M, double *Mp, 
	   double *Mpp){
  /* transition probs
   * adapted from proml make_pmatrix 
   * alpha not used
   */
  static double elambdat[NCHAR];
  double q, p0, p0p, p0pp;
  int i, j, k;

/*   if (t == 0.0){ */
/*     for(i = 0; i < NCHAR; i++) */
/*       for(j = 0; j < NCHAR; j++) M[i + j*NCHAR] = (i == j)?1.0:0.0; */
/*     return; */
/*   } */

  for(k = 0; k < NCHAR; k++) elambdat[k] = exp(t * Lam[k]);
  for(i = 0; i < NCHAR; i++)
    for(j = 0; j < NCHAR; j++){      
      p0 = 0.0; p0p = 0.0; p0pp = 0.0;
      for(k = 0; k < NCHAR; k++){
	q = W[i + k*NCHAR] * W[j+ k*NCHAR];
	p0 += (q * elambdat[k]);
	p0p += (q * Lam[k] * elambdat[k]);
	p0pp += (q * Lam[k] * Lam[k] * elambdat[k]);
      }
      M[i + j*NCHAR] = p0/fr[i];
      Mp[i + j*NCHAR] = p0p/fr[i];
      Mpp[i + j*NCHAR] = p0pp/fr[i];
    }
}
void Mijpd_gam(double t, double *fr, double *W, double *Lam, 
	       double *M, double *Mp, double *Mpp, double alpha){
  static double elambdat[NCHAR];
  double q, p0, p0p, p0pp;
  int i, j, k;

  for(k = 0; k < NCHAR; k++) 
    elambdat[k] = exp(alpha * log(alpha/(alpha - Lam[k] * t)));
  for(i = 0; i < NCHAR; i++)
    for(j = 0; j < NCHAR; j++){      
      p0 = 0.0; p0p = 0.0; p0pp = 0.0;
      for(k = 0; k < NCHAR; k++){
	q = W[i + k*NCHAR] * W[j+ k*NCHAR] * elambdat[k];
	p0 += q;
	p0p += q * alpha * Lam[k] / (alpha - Lam[k]*t);
	p0pp += q * alpha * (alpha + 1.0) * Lam[k] * Lam[k]/
	  ((alpha - Lam[k]*t) * (alpha - Lam[k]*t));
      }
      M[i + j*NCHAR] = p0/fr[i];
      Mp[i + j*NCHAR] = p0p/fr[i];
      Mpp[i + j*NCHAR] = p0pp/fr[i];
    }
}
void WLamfr2exchange(int nchar, double *W, double *Lam, double *fr, double *R){
  int i,j,k;
  for(i = 0; i < nchar; i++)
    for(j = (i+1); j < nchar; j++){
      R[i+j*nchar] = 0.0;
      for(k = 0; k < nchar; k++) 
	R[i+j*nchar] += W[i+k*nchar]*W[j+k*nchar]*Lam[k];
    }
  
  for(i = 0; i < nchar; i++) 
    for(j = (i+1); j < nchar; j++) R[i+j*nchar] /= (fr[i]*fr[j]);
  for(i = 0; i < nchar; i++) 
    for(j = 0; j < i; j++) R[i+j*nchar] = R[j+i*nchar];
  for(i = 0; i < nchar; i++)R[i+i*nchar] = 0.0;
}
void chQ_freq(int nbin, double *W, double *Lam, double *fr, double *fro)
{
  /* binning/udistfbf.c 
   *     Q = Pi^(-1) W Lam W' 
   * decomposition used to obtain P(t) as
   *     P(t) = Pi^(-1) W exp(Lam t) W'
   * where W is determined from eigen-decomp
   *     Pi^(1/2) Q Pi^(-1/2) = U' Lam U,   W = (Pi)^(1/2) U'
   * To adjust for frequencies Pi*:
   * obtain Q_ij* = Q_ij pi_j* / pi_j
   *        Q_ii* = - sum_{j neq j} Q_ij*
   *        M_ij* =  sqrt(pi_i*) Q_ij* / sqrt(pi_j*)
   *        U and Lam from M = U' Lam U (eigen-decomp)
   *        W = (Pi*)^(1/2) U' */
  int i,j,l;
  double R[400], Offdiag[400], Rt, mu;

/*   for(i = 0; i < nbin; i++) */
/*     for(j = 0; j < nbin; j++){ */
/*       Rt = 0.0; */
/*       for(l = 0; l < nbin; l++) */
/* 	Rt += W[i+l*nbin]*W[j+l*nbin]*Lam[l]; */
/*       Rt /= fro[i]; */
/*       if (i != j) */
/* 	printf("%f\n", Rt); */
/*     } */
/*   exit(0); */

  for(i = 0; i < nbin; i++){ /* R_ij = Q_ij / (pi_i pi_j) */
    for(j = (i+1); j < nbin; j++){
      Rt = 0.0;
      for(l = 0; l < nbin; l++){
	Rt += W[i+l*nbin]*W[j+l*nbin]*Lam[l];
      }
      Rt /= (fro[i]*fro[j]);
      R[i+j*nbin] = Rt; R[j+i*nbin] = Rt;
    }
  }

  for(i = 0; i < nbin; i++){ /* R_ii = Q_ii* */
    Rt = 0.0;
    for(j = 0; j < nbin; j++){
      if (j != i)
	Rt -= R[i+j*nbin]*fr[j];
    }
    R[i+i*nbin] = Rt/fr[i]; /* will cancel with fr[i] below */
  }

/*   for(i = 0; i < nbin; i++) */
/*     for(j = 0; j < nbin; j++) printf("%f\n",R[i+j*nbin]*fr[j]); exit(0); */

  for(i = 0; i < nbin; i++){ /* R_ij = entry of M_ij* above */
    for(j = i; j < nbin; j++){
      R[i+j*nbin] *= sqrt(fr[i]*fr[j]);
      if (j != i) R[j+i*nbin] = R[i+j*nbin];
    }
  }

  eigenRealSym(R, nbin, Lam, Offdiag); /* eigen-decomp of M* */
  for(i = 0; i < nbin; i++){ /* W = (Pi*)^(1/2) U' */
    for(j = 0; j < nbin; j++){
      W[i+j*nbin] = R[j+i*nbin] * sqrt(fr[i]);
    }
  }

  /* rescale eigenvalues so that -sum_i pi_i Q_ii = 1 */
  mu = 0.0;
  for(i = 0; i < nbin; i++)
    for(l = 0; l < nbin; l++) mu -= Lam[l]*W[i+l*nbin]*W[i+l*nbin];
  for(i = 0; i < nbin; i++) Lam[i] /= mu;
}
double int_elamrt(double t, double lam, double alpha, int deriv, double *de){
  double te,dt;
  te = -lam*t/alpha;
  dt= exp(-alpha*log(1+te));
  if(deriv==0) return(dt);
  if(deriv==1 || deriv==2){
    de[0]=lam*exp(-(alpha+1)*log(1+te));
    if(deriv==2) de[1]=(alpha+1)*lam*lam*exp(-(alpha+2)*log(1+te))/alpha;
  }
  if(deriv== 3 || deriv==4){
    de[0] = dt*(te/(1+te)-log(1+te));
    if(deriv==4) 
      de[1]=de[0]*de[0]/dt + 
	dt*(te*te/(alpha*(1+te)*(1+te))-te/(alpha*(1+te)));
  }
  return(dt);
}
double pijJCg(int x, int y, double t, int nchar){
  double m;
  m = nchar/(nchar - 1.0);
  if (x == y) 
    return((1.0/nchar) + exp(-t*m)/m);
  else
    return((1.0/nchar) - (1.0/nchar)*exp(-t*m));
}
double pijJCg_gen(int x, int y, double t, int nchar, double *fr, double *W, 
		  double *Lam){
  return(pijJCg(x,y,t,nchar));
}
double ppijJCg(int x, int y, double t, int nchar){ 
  double m;
  m = nchar/(nchar - 1.0);
  if (x == y) 
    return(-exp(-t*m));
  else
    return(exp(-t*m)/(nchar - 1.0));
}
double ppijJCg_gen(int x, int y, double t, int nchar, double *fr,
		   double *W, double *Lam){ 
  return(ppijJCg(x,y,t,nchar));
}
double pp2ijJCg(int x, int y, double t, int nchar){
  double m;
  m = nchar/(nchar - 1.0);
  if (x == y) 
    return(m*exp(-t*m));
  else
    return(-m*exp(-t*m)/(nchar - 1.0));
}
void pJCm(double t, int nchar, double *fr, int deriv, 
	    double *M, double *Mp, double *Mpp){
  int i,j;
  double m,emt;
  m = nchar/(nchar - 1.0);
  emt=exp(-t*m);
  for(i = 0; i < nchar; i++)
    for(j = 0; j < nchar; j++){
      if(i==j){
	M[i+j*nchar]=1.0/nchar+(1.0-1.0/nchar)*emt;
	if(deriv==1 || deriv==2) Mp[i+j*nchar]=-m*(1.0-1.0/nchar)*emt;
	if(deriv==2) Mpp[i+j*nchar]=m*m*(1.0-1.0/nchar)*emt;
      }
      if(i!=j){
	M[i+j*nchar]=1.0/nchar*(1.0-exp(-m*t));
	if(deriv==1 || deriv==2) Mp[i+j*nchar]=m*1.0/nchar*emt;
	if(deriv==2) Mpp[i+j*nchar]=-m*m*1.0/nchar*emt;
      }
    }
}
void pJCm_gen(double t, int nchar, double *fr, double *W, double *Lam, 
		int deriv, double *M, double *Mp, double *Mpp){
  pJCm(t,nchar,fr,deriv,M,Mp,Mpp);
}
void pJCma(double t, int nchar, double *fr, int deriv, double alpha,
	   double *M, double *Mp, double *Mpp){
  int i,j;
  double mu,emut,demu[2];
  mu = nchar/(nchar - 1.0);
  emut=int_elamrt(t,-1.0*mu,alpha,deriv,demu);

  for(i = 0; i < nchar; i++)
    for(j = 0; j < nchar; j++){
      if(i==j){
	M[i+j*nchar]=1.0/nchar+(1.0-1.0/nchar)*emut;
	if(deriv > 0) 
	  Mp[i+j*nchar]=(1.0-1.0/nchar)*demu[0];
	if(deriv==2 || deriv==4) Mpp[i+j*nchar]=(1.0-1.0/nchar)*demu[1];
      }
      if(i!=j){
	M[i+j*nchar]=1.0/nchar-emut/nchar;
	if(deriv > 0) 
	  Mp[i+j*nchar]=-demu[0]/nchar;
	if(deriv==2 || deriv==4) Mpp[i+j*nchar]=-demu[1]/nchar;
      }
    }
}
void pJCma_gen(double t, int nchar, double *fr, double *W, double *Lam, 
	       int deriv, double alpha, double *M, double *Mp, double *Mpp){
  pJCma(t,nchar,fr,deriv,alpha,M,Mp,Mpp);
}
double muF81f(int nchar, double *fr){
  int l;
  double mu = 0.0;
  for(l = 0; l < nchar; l++) mu += fr[l]*(1.0-fr[l]);
  return(1.0/mu);
}
double pijF81(int i, int j, double t, int nchar, double mu, double *fr){
  if (i == j) return(fr[j]+(1.0-fr[j])*exp(-mu*t));
  if (i != j) return(fr[j]*(1.0-exp(-mu*t)));
  return(fr[j]*(1.0-exp(-mu*t)));
}
double pijF81_gen(int i, int j, double t, int nchar, double *fr, double *mu,
		  double *Lam){
  return(pijF81(i,j,t,nchar,*mu,fr));
}
double ppijF81(int i, int j, double t, int nchar, double mu, double *fr){
  if (i == j) return(-mu*(1.0-fr[j])*exp(-mu*t));
  if (i != j) return(mu*exp(-mu*t)*fr[j]);
  return(mu*exp(-mu*t)*fr[j]);
}
double ppijF81_gen(int i, int j, double t, int nchar, double *fr, double *mu,
		   double *Lam){
  return(ppijF81(i,j,t,nchar,*mu,fr));
}
void pF81m(double t, int nchar, double *fr, double mu, int deriv, 
	   double *M, double *Mp, double *Mpp){
  int i,j;
  double emut;
  emut=exp(-mu*t);
  for(i = 0; i < nchar; i++)
    for(j = 0; j < nchar; j++){
      if(i==j){
	M[i+j*nchar]=fr[j]+(1.0-fr[j])*emut;
	if(deriv==1 || deriv==2) Mp[i+j*nchar]=-mu*(1.0-fr[j])*emut;
	if(deriv==2) Mpp[i+j*nchar]=mu*mu*(1.0-fr[j])*emut;
      }
      if(i!=j){
	M[i+j*nchar]=fr[j]*(1.0-exp(-mu*t));
	if(deriv==1 || deriv==2) Mp[i+j*nchar]=mu*fr[j]*emut;
	if(deriv==2) Mpp[i+j*nchar]=-mu*mu*fr[j]*emut;
      }
    }
}
void pF81m_gen(double t, int nchar, double *fr, double *mu, double *Lam, 
		 int deriv, double *M, double *Mp, double *Mpp){
  pF81m(t,nchar,fr,*mu,deriv,M,Mp,Mpp);
}
void pF81ma(double t, int nchar, double *fr, double mu, int deriv, double alpha,
	   double *M, double *Mp, double *Mpp){
  int i,j;
  double emut,demu[2];
  emut=int_elamrt(t,-1.0*mu,alpha,deriv,demu);
  for(i = 0; i < nchar; i++)
    for(j = 0; j < nchar; j++){
      if(i==j){
	M[i+j*nchar]=fr[j]+(1.0-fr[j])*emut;
	if(deriv > 0) Mp[i+j*nchar]=(1-fr[j])*demu[0];
	if(deriv==2 || deriv==4) Mpp[i+j*nchar]=(1-fr[j])*demu[1];
      }
      if(i!=j){
	M[i+j*nchar]=fr[j]-fr[j]*emut;
	if(deriv > 0) Mp[i+j*nchar]=-fr[j]*demu[0];
	if(deriv==2 || deriv==4) Mpp[i+j*nchar]=-fr[j]*demu[1];
      }
    }
}
void pF81ma_gen(double t, int nchar, double *fr, double *mu, double *Lam, 
		int deriv, double alpha, double *M, double *Mp, double *Mpp){
  pF81ma(t,nchar,fr,*mu,deriv,alpha,M,Mp,Mpp);
}
int is_transitionf(int i, int j){
  return((i==0 && j==2) || (i==2 && j==0) || (i==1 && j==3) || (i==3 && j==1));
}
double muF84f(double K, double *fr){
  double mu = 0.0;
  int l;
/*   printf("%f\n",K); */
  for(l = 0; l < 4; l++) mu += fr[l]*(1.0-fr[l]);
  mu += 2*K*( fr[0]*fr[2]/(fr[0]+fr[2]) + fr[1]*fr[3]/(fr[1]+fr[3]) );
  return(1.0/mu);
}
double pijF84(int i, int j, double t, double mu, double K, double *fr){
  double Pj;
  int is_transition;
/*   printf("1.0/(K+1):%f\n",1.0/(K+1)); */
/*   printf("(K+1)*mu*t:%f\n",(K+1)*mu*t); */
/*   printf("mu*t:%f\n",mu*t); */
  if (j == 0 || j == 2) 
    Pj=fr[0]+fr[2];
  else
    Pj=fr[1]+fr[3];

  if(i == j)
    return(fr[j]+fr[j]*(1/Pj - 1)*exp(-mu*t)+(1-fr[j]/Pj)*exp(-mu*(K+1)*t));

  if (i != j){
    is_transition = is_transitionf(i,j);
    if (is_transition)
      return(fr[j]+fr[j]*(1/Pj - 1)*exp(-mu*t)-(fr[j]/Pj)*exp(-mu*(K+1)*t));
    else
      return(fr[j]*(1.0-exp(-mu*t)));
  }
  return(-1.0);
}
double pijF84_gen(int i, int j, double t, int nchar, double *fr, double *mu,
		  double *K){
  return(pijF84(i,j,t,*mu,*K,fr));
}
double ppijF84(int i, int j, double t, double mu, double K, double *fr){
  double Pj,f;
  int is_transition;
  if (j == 0 || j == 2) 
    Pj=fr[0]+fr[2];
  else
    Pj=fr[1]+fr[3];
  
  if(i == j)
    f=-mu*fr[j]*(1/Pj - 1)*exp(-mu*t) + 
      -1.0*mu*(K+1)*(1-fr[j]/Pj)*exp(-mu*(K+1)*t);
  
  if (i != j){
    is_transition = is_transitionf(i,j);
    if (is_transition)
      f = -mu*fr[j]*(1/Pj - 1)*exp(-mu*t)+
	mu*(K+1)*(fr[j]/Pj)*exp(-mu*(K+1)*t);
    else
      f = mu*fr[j]*exp(-mu*t);
  }
  return(f);
}
double ppijF84_gen(int i, int j, double t, int nchar, double *fr, double *mu,
		  double *K){
  return(ppijF84(i,j,t,*mu,*K,fr));
}
double p2ijF84(int i, int j, double t, double mu, double K, double *fr){
  double Pj,f;
  int is_transition;
  if (j == 0 || j == 2) 
    Pj=fr[0]+fr[2];
  else
    Pj=fr[1]+fr[3];
  
  if(i == j)
    f=mu*mu*fr[j]*(1/Pj - 1)*exp(-mu*t) + 
      mu*(K+1)*mu*(K+1)*(1-fr[j]/Pj)*exp(-mu*(K+1)*t);
  
  if (i != j){
    is_transition = is_transitionf(i,j);
    if (is_transition)
      f = mu*mu*fr[j]*(1/Pj - 1)*exp(-mu*t) -
	mu*(K+1)*mu*(K+1)*(fr[j]/Pj)*exp(-mu*(K+1)*t);
    else
      f = -mu*mu*fr[j]*exp(-mu*t);
  }
  return(f);
}
double p2ijF84_gen(int i, int j, double t, int nchar, double *fr, double *mu,
		  double *K){
  return(p2ijF84(i,j,t,*mu,*K,fr));
}
void pF84m(double t, double mu, double K, double *fr, int deriv, 
	   double *M, double *Mp, double *Mpp){
  int i,j;
  double Pr,Py,Pj,muK,emuKt,emut;
  Pr=fr[0]+fr[2];
  Py=fr[1]+fr[3];
  muK=mu*(K+1); emuKt=exp(-muK*t); emut=exp(-mu*t);

  for(j = 0; j < 4; j++){
    if (j == 0 || j == 2) Pj=Pr;
    if (j == 1 || j == 3) Pj=Py;
    for(i = 0; i < 4; i++){
      if(i==j){
	M[i+j*4]=fr[j]+fr[j]*(1/Pj-1)*emut+(1-fr[j]/Pj)*emuKt;
	if(deriv==1 || deriv==2)
	  Mp[i+j*4]=-mu*fr[j]*(1/Pj-1)*emut-muK*(1-fr[j]/Pj)*emuKt;
	if(deriv==2)
	  Mpp[i+j*4]=mu*mu*fr[j]*(1/Pj-1)*emut + muK*muK*(1-fr[j]/Pj)*emuKt;
      }
      if(i!=j){
	if(is_transitionf(i,j)){
	  M[i+j*4]=fr[j]+fr[j]*(1/Pj-1)*emut-(fr[j]/Pj)*emuKt;
	  if(deriv==1 || deriv==2)
	    Mp[i+j*4]=-mu*fr[j]*(1/Pj-1)*emut+muK*(fr[j]/Pj)*emuKt;
	  if(deriv==2)
	    Mpp[i+j*4]=mu*mu*fr[j]*(1/Pj-1)*emut-muK*muK*(fr[j]/Pj)*emuKt;
	}
	else{
	  M[i+j*4]=fr[j]*(1.0-emut);
	  if(deriv==1 || deriv==2) Mp[i+j*4] = mu*fr[j]*emut;
	  if(deriv==2) Mpp[i+j*4]=-mu*mu*fr[j]*emut;
	}
      }
    }
  }
}
void pF84m_gen(double t, int nchar, double *fr, double *mu, double *K, 
	       int deriv, double *M, double *Mp, double *Mpp){
  pF84m(t,*mu,*K,fr,deriv,M,Mp,Mpp);
}
void pF84ma(double t, double mu, double K, double *fr, int deriv, double alpha,
	    double *M, double *Mp, double *Mpp){
  int i,j;
  double Pr,Py,Pj,emuKt,emut,demuK[2],demu[2];
  Pr=fr[0]+fr[2];
  Py=fr[1]+fr[3];
  emut=int_elamrt(t,-1.0*mu,alpha,deriv,demu);
  emuKt=int_elamrt(t,-1.0*mu*(K+1),alpha,deriv,demuK);
/*   if (deriv==4) printf("%f %f\n",emut,emuKt); */

  for(j = 0; j < 4; j++){
    if (j == 0 || j == 2) Pj=Pr;
    if (j == 1 || j == 3) Pj=Py;
    for(i = 0; i < 4; i++){
      if(i==j){
	M[i+j*4]=fr[j]+fr[j]*(1/Pj-1)*emut+(1-fr[j]/Pj)*emuKt;
	if(deriv > 0)
	  Mp[i+j*4]=fr[j]*(1/Pj-1)*demu[0]+(1-fr[j]/Pj)*demuK[0];
	if(deriv==2 || deriv==4)
	  Mpp[i+j*4]=fr[j]*(1/Pj-1)*demu[1]+(1-fr[j]/Pj)*demuK[1];
      }
      if(i!=j){
	if(is_transitionf(i,j)){
	  M[i+j*4]=fr[j]+fr[j]*(1/Pj-1)*emut-(fr[j]/Pj)*emuKt;
	  if(deriv > 0)
	    Mp[i+j*4]=fr[j]*(1/Pj-1)*demu[0]-(fr[j]/Pj)*demuK[0];
	  if(deriv==2 || deriv==4)
	    Mpp[i+j*4]=fr[j]*(1/Pj-1)*demu[1]-(fr[j]/Pj)*demuK[1];
	}
	else{
	  M[i+j*4]=fr[j]-fr[j]*emut;
	  if(deriv > 0) Mp[i+j*4] = -fr[j]*demu[0];
	  if(deriv==2 || deriv==4) Mpp[i+j*4]=-fr[j]*demu[1];
	}
      }
    }
  }
}
void pF84ma_gen(double t, int nchar, double *fr, double *mu, double *K, 
		int deriv, double alpha, double *M, double *Mp, double *Mpp){
  pF84ma(t,*mu,*K,fr,deriv,alpha,M,Mp,Mpp);
}
double tt2K(double R, double *fr){
  double pY,pR,K;
  pY=fr[1]+fr[3]; pR=fr[0]+fr[2];
  K=(R*pR*pY-fr[0]*fr[2]-fr[1]*fr[3])/(fr[0]*fr[2]/pR+fr[1]*fr[3]/pY);
  return(K);
}
double K2tt(double K, double *fr){
  double pY,pR,R;
  pY=fr[1]+fr[3]; pR=fr[0]+fr[2];
  R=K*(fr[0]*fr[2]/pR+fr[1]*fr[3]/pY)/(pR*pY) + 
    (fr[0]*fr[2]+fr[1]*fr[3])/(pR*pY);
  return(R);
}
double muHKYf(double kappa, double *fr){
  double mu;
  mu=2.0*(kappa*(fr[0]*fr[2]+fr[1]*fr[3])+(fr[0]+fr[2])*(fr[1]+fr[3]));
  return(1.0/mu);
}
double pijHKY(int i, int j, double t, double mu, double kappa, double *fr){
  int is_transition;
  double A,Pi_j;
  is_transition=is_transitionf(i,j);
  if (i !=j && !is_transition) 
    return(fr[j]*(1.0-exp(-mu*t)));

  if (j == 0 || j == 2) Pi_j=fr[0]+fr[2];
  if (j == 1 || j == 3) Pi_j=fr[1]+fr[3];
  A = 1+Pi_j*(kappa-1);
  if (i == j){
    return(fr[j]+fr[j]*(1/Pi_j - 1)*exp(-mu*t) +(1-fr[j]/Pi_j)*exp(-mu*t*A));
  }
  /* must be transition */
  return(fr[j]+fr[j]*(1/Pi_j - 1)*exp(-mu*t) -fr[j]*exp(-mu*t*A)/Pi_j);
}
double pijHKY_gen(int i, int j, double t, int nchar, double *fr, double *mu,
		  double *kappa){
  return(pijHKY(i,j,t,*mu,*kappa,fr));
}
double ppijHKY(int i, int j, double t, double mu, double kappa, double *fr){
  int is_transition;
  double A,Pi_j;
  is_transition=is_transitionf(i,j);
  if (i !=j && !is_transition) 
    return(mu*fr[j]*exp(-mu*t));

  if (j == 0 || j == 2) Pi_j=fr[0]+fr[2];
  if (j == 1 || j == 3) Pi_j=fr[1]+fr[3];
  A = 1+Pi_j*(kappa-1);
  if (i == j){
    return(-mu*fr[j]*(1/Pi_j-1)*exp(-mu*t)-mu*A*(1-fr[j]/Pi_j)*exp(-mu*t*A));
  }
  /* must be transition */
  return(-mu*fr[j]*(1/Pi_j - 1)*exp(-mu*t)+mu*A*fr[j]*exp(-mu*t*A)/Pi_j);
}
double ppijHKY_gen(int i, int j, double t, int nchar, double *fr, double *mu,
		   double *kappa){
  return(ppijHKY(i,j,t,*mu,*kappa,fr));
}
double p2ijHKY(int i, int j, double t, double mu, double kappa, double *fr){
  double A,Pi_j;
  if (i !=j && !is_transitionf(i,j)) 
    return(-mu*mu*fr[j]*exp(-mu*t));
  if (j == 0 || j == 2) Pi_j=fr[0]+fr[2];
  if (j == 1 || j == 3) Pi_j=fr[1]+fr[3];
  A = 1+Pi_j*(kappa-1);
  if (i == j){
    return(mu*mu*fr[j]*(1/Pi_j-1)*exp(-mu*t)-mu*A*mu*A*(1-fr[j]/Pi_j)*exp(-mu*t*A));
  }
  /* must be transition */
  return(mu*mu*fr[j]*(1/Pi_j - 1)*exp(-mu*t)-mu*A*mu*A*fr[j]*exp(-mu*t*A)/Pi_j);
}
double p2ijHKY_gen(int i, int j, double t, int nchar, double *fr, double *mu,
		   double *kappa){
  return(p2ijHKY(i,j,t,*mu,*kappa,fr));
}
void pHKYm(double t, int nchar, double *fr, double mu, double kappa,
	   int deriv, double *M, double *Mp, double *Mpp){
  int i,j;
  double Pr,Py,A,emuAt,emut,muA,Pj;

  Pr=fr[0]+fr[2];
  Py=fr[1]+fr[3];
  emut=exp(-mu*t);

  for(j = 0; j < 4; j++){

    if (j == 0 || j == 2) Pj=Pr;
    if (j == 1 || j == 3) Pj=Py;
    A = 1+Pj*(kappa-1); muA=mu*A;
    emuAt=exp(-muA*t);

    for(i = 0; i < 4; i++){
      if(i==j){
	M[i+j*4] = fr[j]+fr[j]*(1/Pj-1)*emut+(1-fr[j]/Pj)*emuAt;
	if(deriv==1 || deriv==2)
	  Mp[i+j*4] = -mu*fr[j]*(1/Pj-1)*emut-muA*(1-fr[j]/Pj)*emuAt;
	if(deriv==2)
	  Mpp[i+j*4] = mu*mu*fr[j]*(1/Pj-1)*emut+muA*muA*(1-fr[j]/Pj)*emuAt;
      }
      if(i!=j){
	if(is_transitionf(i,j)){
	  M[i+j*4] = fr[j]+fr[j]*(1/Pj-1)*emut-fr[j]*emuAt/Pj;
	  if(deriv==1 || deriv==2)
	    Mp[i+j*4] = -mu*fr[j]*(1/Pj-1)*emut+muA*fr[j]*emuAt/Pj;
	  if(deriv==2)
	    Mpp[i+j*4] = mu*mu*fr[j]*(1/Pj-1)*emut-muA*muA*fr[j]*emuAt/Pj;
	}
	else{
	  M[i+j*4] = fr[j]-fr[j]*emut;
	  if(deriv==1 || deriv==2) Mp[i+j*4] = mu*fr[j]*emut;
	  if(deriv==2) Mpp[i+j*4] = -mu*mu*fr[j]*emut;
	}
      }
    }}
}
void pHKYm_gen(double t, int nchar, double *fr, double *mu, double *kappa, 
	       int deriv, double *M, double *Mp, double *Mpp){
  pHKYm(t,nchar,fr,*mu,*kappa,deriv,M,Mp,Mpp);
}
void pHKYma(double t, int nchar, double *fr, double mu, double kappa,
	    int deriv, double alpha, double *M, double *Mp, double *Mpp){
  int i,j;
  double Pr,Py,A,emuAt,demuA[2],emut,demu[2],muA,Pj;

  Pr=fr[0]+fr[2];
  Py=fr[1]+fr[3];
  emut=int_elamrt(t,-1.0*mu,alpha,deriv,demu);

  for(j = 0; j < 4; j++){

    if (j == 0 || j == 2) Pj=Pr;
    if (j == 1 || j == 3) Pj=Py;
    A = 1+Pj*(kappa-1); muA=mu*A;
    emuAt=int_elamrt(t,-1.0*muA,alpha,deriv,demuA);

    for(i = 0; i < 4; i++){
      if(i==j){
	M[i+j*4] = fr[j]+fr[j]*(1/Pj-1)*emut+(1-fr[j]/Pj)*emuAt;
	if(deriv > 0)
	  Mp[i+j*4] = fr[j]*(1/Pj-1)*demu[0]+(1-fr[j]/Pj)*demuA[0];
	if(deriv==2 || deriv==4)
	  Mpp[i+j*4] = fr[j]*(1/Pj-1)*demu[1]+(1-fr[j]/Pj)*demuA[1];
      }
      if(i!=j){
	if(is_transitionf(i,j)){
	  M[i+j*4] = fr[j]+fr[j]*(1/Pj-1)*emut-fr[j]*emuAt/Pj;
	  if(deriv > 0)
	    Mp[i+j*4] = fr[j]*(1/Pj-1)*demu[0]-fr[j]*demuA[0]/Pj;
	  if(deriv==2 || deriv==4)
	    Mpp[i+j*4] = fr[j]*(1/Pj-1)*demu[1]-fr[j]*demuA[1]/Pj;
	}
	else{
	  M[i+j*4] = fr[j]-fr[j]*emut;
	  if(deriv > 0) Mp[i+j*4] = -fr[j]*demu[0];
	  if(deriv==2 || deriv==4) Mpp[i+j*4] = -fr[j]*demu[1];
	}
      }
    }}
}
void pHKYma_gen(double t, int nchar, double *fr, double *mu, double *kappa, 
	       int deriv, double alpha, double *M, double *Mp, double *Mpp){
  pHKYma(t,nchar,fr,*mu,*kappa,deriv,alpha,M,Mp,Mpp);
}
double tt2kappa(double R, double *fr){
  double pR,pY;
  pR=fr[0]+fr[2];
  pY=fr[1]+fr[3];
  return(R*pR*pY/(fr[0]*fr[2]+fr[1]*fr[3]));
}
double kappa2tt(double kappa,  double *fr){
  double pR,pY;
  pR=fr[0]+fr[2];
  pY=fr[1]+fr[3];
  return(kappa*(fr[0]*fr[2]+fr[1]*fr[3])/(pR*pY));
}
void rmnegpv(int m, double *pv, int incp){
  int i,mi;
  double sum=0.0;
  mi=m*incp;
  for(i=0; i<mi; i+=incp)
    if(pv[i] < 0.0)
      pv[i] = 0.0;
    else
      sum += pv[i];
  for(i=0; i<mi; i+=incp) pv[i] /= sum;
}
void rmnegtrans(int m, double *P){
  int i,j;

  for(i=0; i<m; i++)
    for(j=0; j<m; j++)
      if(P[i+j*m] < 0.0){
	rmnegpv(m,&P[i],m);
	break;
      }
  return;
}
void Q2frWLam(int nbin, double *Q, double *fr, double *W, double *Lam){
  /*     Q = Pi^(-1) W Lam W' 
   *  decomposition used to obtain P(t) as
   *     P(t) = Pi^(-1) W exp(Lam t) W'
   * where W is determined from eigen-decomp
   *     Pi^(1/2) Q Pi^(-1/2) = U' Lam U,   W = (Pi)^(1/2) U' 
   * Assumes Q corresponds to GTR model
   */
  int i,j;
  double R[400], Offdiag[400], mu;

  /* pi_j = [1 + sum_i neq j Q_ji/Q_ij]**(-1) */
  for(j = 0; j < nbin; j++){
    fr[j] = 1.0;
    for(i = 0; i < nbin; i++)
      if (j != i) fr[j] += Q[j+i*nbin]/Q[i+j*nbin];
    fr[j] = 1.0 / fr[j];
  }
  for(i = 0; i < nbin; i++)
    for(j = 0; j < nbin; j++) /* R =  Pi^(1/2) Q Pi^(-1/2) */
      R[i+j*nbin] = sqrt(fr[i])*Q[i+j*nbin]/sqrt(fr[j]);

  /* eigen-decomp of Pi^(1/2) Q Pi^(-1/2) */
  eigenRealSym(R, nbin, Lam, Offdiag); 

  for(i = 0; i < nbin; i++){ /* W = (Pi*)^(1/2) U' */
    for(j = 0; j < nbin; j++)
      W[i+j*nbin] = R[j+i*nbin] * sqrt(fr[i]);
  }

  /* rescale eigenvalues so that -sum_i pi_i Q_ii = 1 */
  mu = 0.0;
  for(i = 0; i < nbin; i++)
    for(j = 0; j < nbin; j++) mu -= Lam[j]*W[i+j*nbin]*W[i+j*nbin];
  for(i = 0; i < nbin; i++) Lam[i] /= mu;
}
double pijGTR(int i, int j, double t, int nchar, double *fr, double *W, 
	      double *Lam){
  int k;
  double f;
  if (t == 0.0){
    if (i == j) return(1.0);
    if (i != j) return(0.0);
  }
  
  f=0.0;
  /* printf("%i\n", nchar); */
  for(k = 0; k < nchar; k++){
    /* printf("%f %f\n", W[i + k*nchar], W[j+ k*nchar]); */
    /* printf("%f\n", Lam[k]); */
    f += W[i + k*nchar] * W[j+ k*nchar] * exp(t*Lam[k]);
  }
  return(f/fr[i]);
}
double ppijGTR(int i, int j, double t, int nchar, double *fr, double *W, 
	       double *Lam){
   int k;
   double f;
   f=0.0;
   for(k = 0; k < nchar; k++) 
     f += Lam[k]*W[i + k*nchar] * W[j+ k*nchar] * exp(t*Lam[k]);
   return(f/fr[i]);
}
double p2ijGTR(int i, int j, double t, int nchar, double *fr, double *W, 
	       double *Lam){
   int k;
   double f;
   f=0.0;
   for(k = 0; k < nchar; k++) 
     f += Lam[k]*Lam[k]*W[i + k*nchar] * W[j+ k*nchar] * exp(t*Lam[k]);
   return(f/fr[i]);
}
void pGTRm(double t, int nchar, double *fr, double *W, double *Lam, 
	   int deriv, double *M, double *Mp, double *Mpp){
  int i,j,k;
  double w;
  for(i = 0; i < nchar*nchar; i++) M[i]=0.0;
  if(deriv==1 || deriv==2) for(i = 0; i < nchar*nchar; i++) Mp[i]=0.0;
  if(deriv==2) for(i = 0; i < nchar*nchar; i++) Mpp[i]=0.0;
  
  for(i = 0; i < nchar; i++)
    for(j = 0; j < nchar; j++){
      for(k = 0; k < nchar; k++){
	w=W[i + k*nchar] * W[j+ k*nchar] * exp(Lam[k]*t);
	M[i+j*nchar] += w;
	if(deriv==1 || deriv==2) Mp[i+j*nchar] += Lam[k]*w;
	if(deriv==2) Mpp[i+j*nchar] += Lam[k]*Lam[k]*w;
      }
      M[i+j*nchar] /= fr[i];
      if(deriv==1 || deriv==2) Mp[i+j*nchar] /= fr[i];
      if(deriv==2) Mpp[i+j*nchar] /= fr[i];
    }
  rmnegtrans(nchar,M);
}
void pGTRma(double t, int nchar, double *fr, double *W, double *Lam, 
	    int deriv, double alpha, double *M, double *Mp, double *Mpp){
  int i,j,k;
  double dt,de[2];
  for(i = 0; i < nchar*nchar; i++) M[i]=0.0;
  if(deriv > 0) for(i = 0; i < nchar*nchar; i++) Mp[i]=0.0;
  if(deriv==2 || deriv==4) for(i = 0; i < nchar*nchar; i++) Mpp[i]=0.0;
  
  for(i = 0; i < nchar; i++)
    for(j = 0; j < nchar; j++){
      for(k = 0; k < nchar; k++){
	dt=int_elamrt(t,Lam[k],alpha,deriv,de);
	M[i+j*nchar] += W[i + k*nchar] * W[j+ k*nchar] * dt;
	if(deriv > 0){
	  Mp[i+j*nchar] += W[i + k*nchar] * W[j+ k*nchar] * de[0];
	}
	if(deriv==2 || deriv==4)
	  Mpp[i+j*nchar] += W[i + k*nchar] * W[j+ k*nchar] * de[1];
      }
      M[i+j*nchar] /= fr[i];
      if(deriv > 0) Mp[i+j*nchar] /= fr[i];
      if(deriv==2 || deriv==4) Mpp[i+j*nchar] /= fr[i];
    }
  rmnegtrans(nchar,M);
}
int subst_model_setp(int smodel,
		     double (*(*pij))(int i, int j, double t, int nchar,
				     double *fr, double *W, double *Lam),
		     double (*(*ppij))(int i, int j, double t, int nchar,
				       double *fr, double *W, double *Lam)){
  /* IN:
   * smodel - 0:JC, 1:F81, 2:F84, 3:GTR, 4:PAM, 5:JTT, 6:WAG, 7:mtREV24, 8:HKY,
   *          9:LG
   *
   * OUT:
   * pij,ppij - transition function and first derivative
   */
  switch(smodel){
  case 0:
    *pij=&pijJCg_gen; *ppij=&ppijJCg_gen;
    break;
  case 1:
    *pij=&pijF81_gen; *ppij=&ppijF81_gen;
    break;
  case 2:
    *pij=&pijF84_gen; *ppij=&ppijF84_gen;
    break;
  case 3: case 4: case 5: case 6: case 7: case 9:
    *pij=&pijGTR; *ppij=&ppijGTR;
    break;
  case 8:
    *pij=&pijHKY_gen; *ppij=&ppijHKY_gen;
    break;
  }
  return(1);
}
int subst_model_setpm(int smodel,
		      void (*(*pm))(double t, int nchar, double *fr, 
				    double *W, double *Lam, int deriv, 
				    double *M, double *Mp, 
				    double *Mpp)){
  /* IN:
   * smodel - 0:JC, 1:F81, 2:F84, 3:GTR, 4:PAM, 5:JTT, 6:WAG, 7:mtREV24, 8:HKY,
   *          9:LG
   *
   * OUT:
   * pm - transition matrix function 
   */
  switch(smodel){
  case 0:
    *pm=&pJCm_gen; 
    break;
  case 1:
    *pm=&pF81m_gen; 
    break;
  case 2:
    *pm=&pF84m_gen; ;
    break;
  case 3: case 4: case 5: case 6: case 7: case 9:
    *pm=&pGTRm; 
    break;
  case 8:
    *pm=&pHKYm_gen;
    break;
  }
  return(0);
}
int subst_model_setpma(int smodel,
		       void (*(*pm))(double t, int nchar, double *fr, 
				     double *W, double *Lam, int deriv, 
				     double alpha,
				     double *M, double *Mp, double *Mpp)){
  /* IN:
   * smodel - 0:JC, 1:F81, 2:F84, 3:GTR, 4:PAM, 5:JTT, 6:WAG, 7:mtREV24, 8:HKY,
   *          9:LG
   *
   * OUT:
   * pm - transition matrix function 
   */
  switch(smodel){
  case 0:
    *pm=&pJCma_gen; 
    break;
  case 1:
    *pm=&pF81ma_gen; 
    break;
  case 2:
    *pm=&pF84ma_gen; ;
    break;
  case 3: case 4: case 5: case 6: case 7: case 9:
    *pm=&pGTRma; 
    break;
  case 8:
    *pm=&pHKYma_gen;
    break;
  }
  return(0);
}
int subst_model_setparam(int nchar, int smodel, double *fr, double *Qk,
			 double **W, double **Lam){
  /* IN:
   * nchar - number of character states
   * smodel - 0:JC, 1:F81, 2:F84, 3:GTR, 4:PAM, 5:JTT, 6:WAG, 7:mtREV24, 8:HKY,
   *          9:LG
   * Qk - rate matrix for GTR, ttratio, K or Kappa for F84, HKY, < 0 => ttratio; 
   * not used otherwise. 
   *
   * IN/OUT:
   * fr - frequencies; reset for JC, determined from Q for GTR
   *      if *fr < 0 for aa models, model-based frequencies used
   * OUT:
   * W,Lam - contain parameters for substitution model
   *
   * RETURN:
   * 2,1 or 0 according to whether both W and Lam were allocated, just W or
   * neither
   */
  int allocWLam=0,aamodel=0,i;
  double *fro,*Wo,*Lamo;
  if (smodel==4 || smodel==5 || smodel==6 || smodel==7 || smodel==9) aamodel=1;

  if (smodel==0){ /* JC */
    for(i = 0; i < nchar; i++) fr[i] = 1.0/nchar;
  }
  if (smodel==1){ /* F81 */
    *W=(double *) malloc((size_t) sizeof(double));
    **W=muF81f(nchar,fr);
/*     for(i = 0; i < nchar; i++) printf("%f\n",fr[i]); printf("\n"); */
/*     printf("%f\n",**W); */
    allocWLam=1;
  }
  if (smodel==2){ /* F84 */
    *W=(double *) malloc((size_t) sizeof(double));
    *Lam=(double *) malloc((size_t) sizeof(double));
/*     printf("%.16e\n",*Qk); */
/*     for(i = 0; i < nchar; i++) printf("%f\n",fr[i]); printf("\n"); */
    **Lam=*Qk;
    if(*Qk < 0) 
      **Lam=tt2K(-(*Qk),fr);
    **W= muF84f(**Lam,fr);
/*     printf("%f %f\n",**W,**Lam); */
    allocWLam=2;
  }
  if (smodel==3){ /* GTR */
    *W = (double *) malloc((size_t) nchar*nchar*sizeof(double));
    *Lam = (double *) malloc((size_t) nchar*sizeof(double));
    Q2frWLam(nchar,Qk,fr,*W,*Lam);
    allocWLam=2;
  }
  if (smodel==4){ /* PAM */
    Wo = pamprobmat; Lamo = pameigmat;
  }
  if (smodel==5){ /* JTT */
    Wo = jttprobmat; Lamo = jtteigmat;
  }
  if (smodel==6){ /* WAG */
    Wo = wagprobmat; Lamo = wageigmat;
  }
  if (smodel==7){ /* mtREV24 */
    Wo = mtREV24probmat; Lamo = mtREV24eigmat;
  }
  if (smodel==8){ /* HKY */
    *W=(double *) malloc((size_t) sizeof(double));
    *Lam=(double *) malloc((size_t) sizeof(double));
    **Lam=*Qk;
    if(*Qk < 0) 
      **Lam=tt2kappa(-(*Qk),fr);
    **W = muHKYf(**Lam,fr);
    allocWLam=2;
  }
  if(smodel==9){ /* LG */
    Wo = lgprobmat; Lamo = lgeigmat;
  }
  if (aamodel){  /* amino acid models */
    if (*fr > 0){ /* different frequencies than for model */
      *W = (double *) malloc((size_t) nchar*nchar*sizeof(double));
      *Lam = (double *) malloc((size_t) nchar*sizeof(double));
      fro = (double *) malloc((size_t) nchar*sizeof(double));
      makeprotfreqs(fro,Wo,Lamo);
      memcpy(*W,Wo,(size_t) nchar*nchar*sizeof(double));
      memcpy(*Lam,Lamo,(size_t) nchar*sizeof(double));
      chQ_freq(nchar,*W,*Lam,fr,fro);
      allocWLam=2;
    }
    else{
      *W=Wo; *Lam=Lamo;
      makeprotfreqs(fr,*W,*Lam);
    }
  }
  return(allocWLam);
}
int subst_model_set(int nchar, int smodel, double *fr, double *Qk, 
		    double (*(*pij))(int i, int j, double t, int nchar, 
				     double *fr, double *W, double *Lam),
		    double (*(*ppij))(int i, int j, double t, int nchar, 
				      double *fr, double *W, double *Lam),
		    double **W, double **Lam)
{
  /* IN:
   * nchar - number of character states
   * smodel - 0:JC, 1:F81, 2:F84, 3:GTR, 4:PAM, 5:JTT, 6:WAG, 7:mtREV24, 8:HKY,
   *          9:LG
   * Qk - rate matrix for GTR, K or Kappa for F84, HKY; not used otherwise
   *
   * IN/OUT:
   * fr - frequencies; reset for JC
   *      if *fr < 0 for aa models, model-based frequencies used
   * OUT:
   * pij,ppij - transition function and first derivative
   * W,Lam - contain parameters for substitution model
   *
   * RETURN:
   * 2,1 or 0 according to whether both W and Lam were allocated, just W or 
   * neither
   */
  subst_model_setp(smodel,pij,ppij);
  return(subst_model_setparam(nchar,smodel,fr,Qk,W,Lam));
}
int subst_model_setm(int nchar, int smodel, double *fr, double *Qk, 
		     void (*(*pm))(double t, int nchar, double *fr, 
				   double *W, double *Lam, int deriv, 
				   double *M, double *Mp, 
				   double *Mpp),
		     double **W, double **Lam)
{
  /* IN:
   * nchar - number of character states
   * smodel - 0:JC, 1:F81, 2:F84, 3:GTR, 4:PAM, 5:JTT, 6:WAG, 7:mtREV24, 8:HKY,
   *          9:LG
   * Qk - rate matrix for GTR, K or Kappa for F84, HKY, < 0 => ttratio; 
   * not used otherwise. 
   *
   * IN/OUT:
   * fr - frequencies; reset for JC, determined from Q for GTR
   *      if *fr < 0 for aa models, model-based frequencies used
   * OUT:
   * pm - transition matrix function 
   * W,Lam - contain parameters for substitution model
   *
   * RETURN:
   * 2,1 or 0 according to whether both W and Lam were allocated, just W or 
   * neither
   */
  subst_model_setpm(smodel,pm);
  return(subst_model_setparam(nchar,smodel,fr,Qk,W,Lam));
}
void trans_gen(int nchar, int ntaxa, double *utreec, double *fr, double *W, 
	       double *Lam, 		      
	       double pij(int i, int j, double t, int nchar, 
			  double *fr, double *W, double *Lam),
	       double ppij(int i, int j, double t, int nchar, 
			     double *fr, double *W, double *Lam),
	       double *M, double *Mp)
{
  /* IN:
   * utreec: the tree, nutreec format, 0 last edge
   * OUT:
   * M, Mp - transition matrices and first derivatives for each of the edges
   */
  int nchar2,join,kup,k,r,x,y;
  double tr;

  nchar2 = nchar*nchar;
  for(join = 0; join < ntaxa-1; join++){
    kup = (join < ntaxa-2)?2:1;
    for(k = 0; k < kup; k++){
      r = (int) utreec[k+join*4];
      tr = utreec[k+2+join*4];
      for(x = 0; x < nchar; x++)
	for(y = 0; y < nchar; y++){
	  M[x+y*nchar+r*nchar2]=pij(x,y,tr,nchar,fr,W,Lam);
	  Mp[x+y*nchar+r*nchar2]=ppij(x,y,tr,nchar,fr,W,Lam);
	}
      /* printf("r: %i, tr: %f\n", r, tr); */
      /* wmat(stdout,nchar,nchar,&M[r*nchar2]); */
/*       wmat(stdout,nchar,nchar,&M[r*nchar2]); */
      /* wmat(stdout,nchar,nchar,&Mp[r*nchar2]); */
    }
  }
}
void trans_gen_rate(int nchar, int ntaxa, double *utreec, 
		    int nrate, double *rate, double *wt,
		    double *fr, double *W, double *Lam, 		      
		    double pij(int i, int j, double t, int nchar, 
			       double *fr, double *W, double *Lam),
		    double ppij(int i, int j, double t, int nchar, 
				double *fr, double *W, double *Lam),
		    double *M, double *Mp)
{
  /* IN:
   * utreec: the tree, nutreec format, 0 last edge
   * OUT:
   * M, Mp - transition matrices and first derivatives for each edges x rate
   *         nchar x nchar x nbranch x nrate 
   *         derivatives are not multiplied by rate
   */
  int nchar2,nb,j,l,i;
  double *utreecn;
  
  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  utreecn = (double *) malloc((size_t) (ntaxa-1)*4*sizeof(double));
  memcpy(utreecn,utreec,(ntaxa-1)*4*sizeof(double));
  for(j = 0; j < nrate; j++){
    for(i = 0; i < ntaxa-1; i++)
      for(l = 0; l < 2; l++) utreecn[l+2+i*4] = rate[j]*utreec[l+2+i*4];
    trans_gen(nchar,ntaxa,utreecn,fr,W,Lam,pij,ppij,&M[j*nchar2*nb],
	      &Mp[j*nchar2*nb]);
  }
  free(utreecn);
  return;
}
void transm_rate(int smodel, int nchar, int ntaxa, double *utreec, 
		 int nrate, double *rate, double alpha,
		 double *fr, double *W, double *Lam, int deriv, 
		 double *M, double *Mp, double *Mpp){
  int nchar2,nb,j,join,kup,r,i,k;
  double tr,one,*Mpt,*Mppt;
  void (*pm)(double t, int nchar, double *fr, double *W, double *Lam, 
	     int deriv, double *M, double *Mp, double *Mpp);
  void (*pma)(double t, int nchar, double *fr, double *W, double *Lam, 
	      int deriv, double alpha, double *M, double *Mp, double *Mpp);
  if (alpha < 0)
    subst_model_setpm(smodel,&pm);
  if (alpha > 0) 
    subst_model_setpma(smodel,&pma);
  nchar2 = nchar*nchar;
  nb = 2*ntaxa-3;
  for(j = 0; j < nrate; j++){
    for(join = 0; join < ntaxa-1; join++){
      kup = (join < ntaxa-2)?2:1;
      for(k = 0; k < kup; k++){
	r = (int) utreec[k+join*4];
	tr = utreec[k+2+join*4]*rate[j];
	Mpt=&one; Mppt=&one;
	if(deriv > 0) Mpt=&Mp[r*nchar2+j*nchar2*nb];
	if(deriv==2||deriv==4) Mppt=&Mpp[r*nchar2+j*nchar2*nb];
	if(alpha < 0)
	  (*pm)(tr,nchar,fr,W,Lam,deriv,&M[r*nchar2+j*nchar2*nb],Mpt,Mppt);
	if(alpha > 0)
	  (*pma)(tr,nchar,fr,W,Lam,deriv,alpha,&M[r*nchar2+j*nchar2*nb],
		 Mpt,Mppt);	
	if(deriv==1 || deriv==2)
	  for(i = 0; i < nchar2; i++) 
	    Mp[i+r*nchar2+j*nchar2*nb] *= rate[j];
	if(deriv==2)
	  for(i = 0; i < nchar2; i++) 
	    Mpp[i+r*nchar2+j*nchar2*nb] *= rate[j]*rate[j];
      }}}
}
int l2i(char c){
  if (c == 'a' || c == 'A') return(0);
  if (c == 'c' || c == 'C') return(1);
  if (c == 'g' || c == 'G') return(2);
  if (c == 't' || c == 'T') return(3);
  if (c == '-') return(5);
  return(6);
}
char i2l(int i){
  switch(i){
  case 0:
    return('A');
  case 1:
    return('C');
  case 2:
    return('G');
  case 3:
    return('T');
  }
  return('E');
}
char i2lp(int i){
  switch(i){
  case 0:
    return('A');

  case 1:
    return('R');
  
  case 2:
    return('N');
 
  case 3:
    return('D');
 
  case 4:
    return('C');
 
  case 5:
    return('Q');
 
  case 6:
    return('E');
 
  case 7:
    return('G');
 
  case 8:
    return('H');
 
  case 9:
    return('I');
 
  case 10:
    return('L');
 
  case 11:
    return('K');
 
  case 12:
    return('M');
 
  case 13:
    return('F');
 
  case 14:
    return('P');
 
  case 15:
    return('S');
 
  case 16:
    return('T');
 
  case 17:
    return('W');
 
  case 18:
    return('Y');
 
  case 19:
    return('V');
    
  case 21:
    return('-');
    
  case 22:
    return('X');
    
  case 23:
    return('?');
  }
  return('Z');
}
int l2ip(char c){
  switch(c){
  case 'A':
    return(alanine);
    
  case 'R':
    return(arginine);
    
  case 'N':
    return(asparagine);
    
  case 'D':
    return(aspartic);

  case 'C':
    return(cysteine);

  case 'Q':
    return(glutamine);

  case 'E':
    return(glutamic);

  case 'G':
    return(glycine);

  case 'H':
    return(histidine);

  case 'I':
    return(isoleucine);

  case 'L':
    return(leucine);

  case 'K':
    return(lysine);

  case 'M':
    return(methionine);

  case 'F':
    return(phenylalanine);

  case 'P':
    return(proline);

  case 'S':
    return(serine);

  case 'T':
    return(threonine);

  case 'W':
    return(tryptophan);

  case 'Y':
    return(tyrosine);

  case 'V':
    return(valine);

  case '-':
    return(21);
    
  case 'X':
    return(22);

  case '?':
    return(23);
    /* future: cases X and ? represent unknown aa 
     * B - asparagine or aspartic
     * Z - glutamine or glutamic
     * * - stop codon
     */
  }
  return(100);
}
void rmgap(int numsp, int *nsite, char *seqc){
  int i,j,k,sp,newnsite;

  newnsite = *nsite;
  for(i = *nsite-1; i >= 0; i--)
    for(j = 0; j < numsp; j++)
      if (seqc[i + j*(*nsite)] == '-'){
	for(k = i; k < newnsite-1; k++)
	  for(sp = 0; sp < numsp; sp++)
	    seqc[k + sp*(*nsite)] = seqc[k+1 + sp*(*nsite)];
	newnsite--;
	break;
      }

  for(j = 1; j < numsp; j++)
    for(i = 0; i < newnsite; i++)
      seqc[i + j*newnsite] = seqc[i + j*(*nsite)];

  *nsite = newnsite;
}
int rmgap_pair(int nsite, int *s1, int *s2, int *seq1, int *seq2)
{
  int ns = 0, i;

  for(i = 0; i < nsite; i++){
    if (s1[i] < 20 && s2[i] < 20){
      seq1[ns] = s1[i];
      seq2[ns] = s2[i];
      ns++;
    }}
  return(ns);
}
int rseqf(FILE *fp, char *seq){
  int rsite = 0;
  char c;
  
  while((c = getc(fp)) != '\n' &&  c != EOF){
    if (c != ' ' && c != '\r' && c!= '\t'){
      *seq = c;
      seq++;
      rsite++;
    }
  }

  return(rsite);
}
char ignore_whitespace_seqfile(FILE *infile){
  char c;
  while((c = getc(infile)) == ' ' || c == '\n' || c == '\r' || c == '\t') ;
  if(c == EOF){
    printf("seqfile: EOF before entire sequence read");
    exit(1);
  }
  return(c);
}
int rinterleave_block(FILE *infile, int numsp, int nsite, char name[][11], 
		       char *seq, int *initb){
  char c;
  int i,j,lsite,lsiteo;
  
  /* printf("%i\n",numsp); */
  for(i = 0; i < numsp; i++){
    c = ignore_whitespace_seqfile(infile);
    if((*initb) == 1){
      name[i][0] = c;
      for(j = 1; j < 10; j++)name[i][j] = getc(infile);
      name[i][10] = '\0';
      c = ignore_whitespace_seqfile(infile);      
      /* printf("%s\n",name[i]); */
    }
    seq[i*nsite] = c;
    lsite = rseqf(infile,&seq[1+i*nsite])+1;
    /* printf("%i\n",lsite); */
    if(i==0) lsiteo=lsite;

    if(i > 0 && lsite != lsiteo){
      printf("seqfile: number of character states should be same for all taxa in a block\n");
      exit(1);
    }
  }
  *initb = 0;
  
  return(lsite);
}
void rinterleavef_nolim(FILE *infile, const int *numsp, const int *nsite, 
			char name[][11], char *seq)
{
  int lsite = 0, csite = 0, initb = 1;

  while(csite < *nsite){
    lsite = rinterleave_block(infile,*numsp,*nsite,name,&seq[csite],&initb);
    csite += lsite;
  }
}
void rinterleave(int *numsp, int *nsite, char name[][11], char *seq)
{
  int fso;
  fso=scanf("%i %i", numsp, nsite); 
  rinterleavef_nolim(stdin,numsp,nsite,name,seq);
}
void rinterleavef(FILE *infile, int *numsp, int *nsite, char name[][11], 
		  char *seq)
{
  int fso;
  fso=fscanf(infile, "%i %i", numsp, nsite); 
  rinterleavef_nolim(infile,numsp,nsite,name,seq);
}
void letter(int numsp, int nsite, char *seq)
{
  int i, j;
  for(i = 1; i < numsp; i++)
    for(j = 0; j < nsite; j++)
      if (seq[j + i*nsite] == '.') seq[j + i*nsite]=seq[j];
}
void pchar_freq(int nchar, int nsite, int *s1, int *s2, double *pf)
{
  int i,j,nc=0;

  for(j = 0; j < nchar*nchar; j++) pf[j] = 0.0;

  for(i = 0; i < nsite; i++){
    if(s1[i] < nchar && s2[i] < nchar){
      pf[s1[i]+s2[i]*nchar]++;
      nc++;
    }
  }

  for(j = 0; j < nchar*nchar; j++) pf[j] /= nc;
}
int pchar_freqi(int nchar, int nsite, int *s1, int *s2, int *nxy)
{
  int i,n=0;

  for(i = 0; i < nchar*nchar; i++) nxy[i] = 0.0;

  for(i = 0; i < nsite; i++){
    if(s1[i] < nchar && s2[i] < nchar){
      nxy[s1[i]+s2[i]*nchar]++;
      n++;
    }
  }
  return(n);
}
void char_freq(int nchar, int numsp, int nsite, int *seq, double *freq)
{
  int i,j,nc=0;

  for(j = 0; j < nchar; j++){
    freq[j] = 0.0;
  }
  for(i = 0; i < nsite*numsp; i++){
    if(seq[i] < nchar){
      for(j = 0; j < nchar; j++){
	if (seq[i] == j) freq[j]++;
      }
      nc++;
    }
  }

  for(j = 0; j < nchar; j++){
    freq[j] /= nc;
  }
}
void char_freq_count(int nchar, int numsp, int endsite, int *seq, int *count, 
		     double *freq)
{
  int i,j,nc=0,k;

  for(j = 0; j < nchar; j++){
    freq[j] = 0.0;
  }

  for(i = 0; i < endsite; i++){
    for(k = 0; k < numsp; k++)
      if(seq[i+k*endsite] < nchar)
	for(j = 0; j < nchar; j++)
	  if (seq[i+k*endsite] == j){
	    freq[j] += count[i];
	    nc += count[i];
	  }
  }

  for(j = 0; j < nchar; j++) freq[j] /= nc;
}
int patt_freq(int numsp, int nsite, int *seq, int *count)
{
  int *s;
  int endsite,i,j,k,match;

  s = (int *) malloc((size_t) nsite*numsp*sizeof(int));
  endsite = 0;
  for(i = 0; i < nsite; i++){
    /* check for a match */
    match = 0;
    for(k = 0; k < endsite; k++){
      match = 1;
      for(j = 0; j < numsp; j++){
	if (seq[i+j*nsite] != s[j+k*numsp]){
	  match = 0;
	  break;
	}
      }
      if (match){
	count[k] += 1;
	break;
      }
    }
    
    /* if not a match, new element */
    if (!match){
      for(j = 0; j < numsp; j++) s[j+endsite*numsp] = seq[i+j*nsite];
      count[endsite] = 1;
      endsite++;
    }
  }
  

  for(j = 0; j < numsp; j++)
    for(i = 0; i < endsite; i++)
      seq[i+j*endsite] = s[j+i*numsp];

  free(s);
  return(endsite);
}
int patt_freq_locs(int numsp, int nsite, int *seq, int *count, int *locs)
{
  /* locs[i] gives the location of ith site in compressed seq matrix */
  int *s;
  int endsite,i,j,k,match;

  s = (int *) malloc((size_t) nsite*numsp*sizeof(int));
  for(i = 0; i < nsite; i++) count[i]=0;
  endsite = 0;
  for(i = 0; i < nsite; i++){
    /* check for a match */
    match = 0;
    for(k = 0; k < endsite; k++){
      match = 1;
      for(j = 0; j < numsp; j++){
	if (seq[i+j*nsite] != s[j+k*numsp]){
	  match = 0;
	  break;
	}
      }
      if (match){
	count[k] += 1;
	locs[i]=k;
	break;
      }
    }
    
    /* if not a match, new element */
    if (!match){
      for(j = 0; j < numsp; j++) s[j+endsite*numsp] = seq[i+j*nsite];
      count[endsite] = 1;
      locs[i]=endsite;
      endsite++;
    }
  }
  

  for(j = 0; j < numsp; j++)
    for(i = 0; i < endsite; i++)
      seq[i+j*endsite] = s[j+i*numsp];

  free(s);
  return(endsite);
}
int all_sitepatt(int nchar, int ntaxa, int *patt)
{
  /* IN:
   * patt: nchar^ntaxa x ntaxa matrix
   *       each row gives a pattern
   * RETURN: nchar^ntaxa
   *
   * updates patt as 
   *            patt 0
   *            patt 1
   *            ...
   *            patt nchar
   */
  int npatt,npattf,i,j,b,l;
  
  npattf = 1;
  for(i = 0; i < ntaxa; i++) npattf *= nchar;
  for(b = 0; b < nchar; b++) patt[b] = b;
  npatt = nchar;
  for(i = 1; i < ntaxa; i++){

    for(j = 0; j <= i-1; j++)
      for(b = 0; b < nchar; b++)
	for(l = 0; l < npatt; l++){
	  patt[l+b*npatt+j*npattf] = patt[l+j*npattf];
	}

    for(b = 0; b < nchar; b++)
      for(l = 0; l < npatt; l++){ 
	patt[l+b*npatt+i*npattf] = b;
      }

    npatt *= nchar;
  }

  return(npattf);
}
int all_sitepattr(int nchar, int ntaxa, int *patt){
  /* IN/OUT:
   * patt: ntaxa vector; IN: current pattern, OUT: next pattern
   *       set to 0 initially
   * RETURN: 0 if final pattern input, 1 otherwise */
  int i,j;
  for(i = 0; i < ntaxa; i++)
    if(patt[i] != nchar-1){
      patt[i] += 1;
      for(j=i-1; j >= 0; j--) patt[j]=0;
      return(1);
    }
  return(0);
}
void pr_utreec_top(FILE *fp, int ntaxa, int *utreec){
  int i;
  for(i = 0; i < ntaxa-1; i++) 
    fprintf(fp, "%i %i\n", utreec[i*2], utreec[1+i*2]);
}
void rlord_utreec(int ntaxa, double *utreec){
  int i;
  double c;
  for(i = 0; i < ntaxa-1; i++)
    if (utreec[i*4] > utreec[1+i*4]){
      c = utreec[1+i*4];
      utreec[1+i*4] = utreec[i*4];
      utreec[i*4] = c;
      c = utreec[3+i*4];
      utreec[3+i*4] = utreec[2+i*4];
      utreec[2+i*4] = c;
    }
}
void utreec2bl(int ntaxa, double *utreec, double *bl){
  int i,j;
/*   pr_utreec(stdout,ntaxa,utreec); */
  for(i = 0; i < ntaxa-2; i++)
    for(j = 0; j < 2; j++){
/*       printf("%i %f\n",(int) utreec[j+i*4],utreec[j+2+i*4]); */
      bl[(int) utreec[j+i*4]] = utreec[j+2+i*4];
    }
/*   printf("%i %f\n",(int) utreec[(ntaxa-2)*4],utreec[2+(ntaxa-2)*4]); */
  bl[(int) utreec[(ntaxa-2)*4]] = utreec[2+(ntaxa-2)*4]+utreec[3+(ntaxa-2)*4];
}
void bl2utreec(int ntaxa, double *bl, double *utreec){
  int i,j;
  for(i = 0; i < ntaxa-2; i++)
    for(j = 0; j < 2; j++)
      utreec[j+2+i*4] = bl[(int) utreec[j+i*4]];
  utreec[2+(ntaxa-2)*4] = bl[(int) utreec[(ntaxa-2)*4]];
  utreec[3+(ntaxa-2)*4] = 0;
}
void pr_distmat(FILE *fp, int numsp, char name[][11], double *dmat){
  /* only need upper triangle */
  int p1,p2;
  fprintf(fp, "%i\n", numsp);
  for(p1 = 0; p1 < numsp; p1++){
    fprintf(fp, "%s ", name[p1]);
    for(p2 = 0; p2 < p1; p2++) fprintf(fp, "%f ", dmat[p2 + p1*numsp]);
    fprintf(fp, "%f ", 0.0);
    for(p2 = p1+1; p2 < numsp; p2++) fprintf(fp, "%f ", dmat[p1 + p2*numsp]);
    fprintf(fp, "\n");
  }
}
void infilefreq(int nbin, int ntaxa, int nsite, int *seq, double *freq)
{
  /* binning/udistfbf.c */
  int i,j,nt;

  /* printf("%i nbin",nbin); */
  nt = 0;
  for(j = 0; j < nbin; j++) freq[j] = 0.0;

  for(i = 0; i < nsite*ntaxa; i++)
    for(j = 0; j < nbin; j++)
      if (seq[i] == j){
	freq[j]++;
	nt++;
	break;
      }

  for(j = 0; j < nbin; j++) freq[j] /= nt;
}
char *read_seqc(FILE *fp, int *ntaxa, int *nsite, char (**name)[11]){
  char *seqc;
  int fso;

  fso=fscanf(fp, "%i %i",ntaxa,nsite); 
  *name=(char (*)[11]) malloc((size_t) (*ntaxa)*sizeof(**name));
  seqc = (char *) malloc((size_t) (*ntaxa)*(*nsite)*sizeof(char));
  rinterleavef_nolim(fp,ntaxa,nsite,*name,seqc);
  return(seqc);
}
int *read_seq(FILE *fp, int nchar, int rm_gap, int *ntaxa, int *nsite, 
	      char (**name)[11]){
  int i,*seq;
  char *seqc;
  
  seqc = read_seqc(fp,ntaxa,nsite,name);
  letter(*ntaxa,*nsite,seqc);
  if (rm_gap) rmgap(*ntaxa,nsite,seqc);
  seq = (int *) malloc((size_t) (*ntaxa)*(*nsite)*sizeof(int));
  if(nchar==20)
    for(i = 0; i < (*nsite)*(*ntaxa); i++) seq[i] = l2ip(seqc[i]);
  else
    for(i = 0; i < (*nsite)*(*ntaxa); i++) seq[i] = l2i(seqc[i]);
  free(seqc);
  return(seq);
}
int names_infile(FILE *infile, char (**names)[11], int sequential){
  /* Will often work with sequential=0 even if file actually is sequential
   * If !sequential
   *   lines 2 through ntaxa+1 start with taxa names 
   *  PHYLIP format allows
   *  Archaeopt
   *  0011001101
   * if sequential option given 
   *
   * names allocated as a contiguous block */
  char c;
  int i,j,nchar,nsite,ntaxa,fso;

  fso=fscanf(infile,"%i %i",&ntaxa,&nsite);
  *names=(char (*)[11]) malloc((size_t) ntaxa*sizeof(**names));
  while((c=getc(infile)) !='\n') ;
  if (!sequential)
    for(j=0; j<ntaxa; j++){
      for(i=0; i<10; i++)
	(*names)[j][i]=getc(infile);
      (*names)[j][10]='\0';
      while((c=getc(infile)) !='\n' && c != EOF) ;
    }
  else{
    for(j=0; j<ntaxa; j++){
      for(i=0; i<10; i++) (*names)[j][i]=getc(infile);
      nchar=0;
      while((c=getc(infile)) != ' ' && c != '\t' && c != '\r' && c != '\n'){
	nchar++;
	if (nchar == nsite) break;
      }
      while((c=getc(infile)) !='\n') ;
    } /* end of for(j=0... */
  } /* end of else */
  return(ntaxa);
}
void parent_branch(int r, int ntaxa, double *utreec, int *p, int *l, 
		   double *bl, double *br)
{
  int i,k,done=0;
  i=r-ntaxa+1; /* only need to check above r */
  if (i < 0) i=0;
  while(!done){
    for(k = 0; k < 2; k++)
      if(((int) utreec[k+i*4])==r){
	*p=i+ntaxa;
	*br=utreec[k+2+i*4];
	*l=utreec[1-k+i*4];
	*bl=utreec[1-k+2+i*4];
	done=1;
      }
    i++;
  }
}
void relab_int(int ntaxa, double *utreec, double *nutreec)
{
  /* assumes r < l */
  int l,la,j,k,r;
 
  for(l = 0; l < 4; l++) nutreec[l+(ntaxa-2)*4] = utreec[l+(ntaxa-2)*4];
  la=2*ntaxa-3;
  /* printf("%i %i\n", (int) nutreec[(ntaxa-2)*4], (int) nutreec[1+(ntaxa-2)*4]); */
  
  for(j=(ntaxa-2); j >= 0; j--)
    for(k = 1; k >= 0; k--){
      r=nutreec[k+j*4];
      if (r >= ntaxa){
	for(l = 0; l < 4; l++) 
	  nutreec[l+(la-ntaxa)*4]=utreec[l+(r-ntaxa)*4];
	nutreec[k+j*4]=la;
	/* printf("j: %i %i\n",(int) nutreec[(ntaxa-2)*4],(int) nutreec[1+(ntaxa-2)*4]); */
	la--;
      }}
}
void rert(int r, int ntaxa, double *utreec, double *nutreec)
{
  /* IN:
   * r - label of branch tree is to be rooted at
   * ntaxa, utreec
   * OUT:
   * nutreec
   *
   * see software notes 2009-06-26
   */
  int p,l,k,kup,j,c,nc;
  double br,bl;
  
  /* If parent node of r is the root, no re-rooting is required. */
  parent_branch(r,ntaxa,utreec,&p,&l,&bl,&br);
  if (p == 2*ntaxa-2){
    for(j = 0; j < 4*(ntaxa-1); j++) nutreec[j]=utreec[j];
    return;
  }

  /* The last row of nutreec will be r 2*ntaxa-3; edge length split in half */
  c=ntaxa-2; 
  nutreec[c*4]=r; nutreec[1+c*4]=2*ntaxa-3;
  nutreec[2+c*4]=br/2; nutreec[3+c*4]=br/2;
  /* printf("%i %i %i\n", c, (int) nutreec[c*4], (int) nutreec[1+c*4]); */

  /* As long as the parent of r is not the root, create new joins in 
   * nutreec that join l with p where p is the parent of l r. 
   * This effectively reverses the orientation of p */
  while(p != 2*ntaxa-2){
    c--; 
    /* The relabeled p will be ntaxa+c-1 rather than l. 
     * New backward branchs start labeling from the root and decrease. */
    nutreec[c*4]=l; nutreec[1+c*4]=ntaxa+c-1;
    nutreec[2+c*4]=bl;
    
    r=p;
    parent_branch(r,ntaxa,utreec,&p,&l,&bl,&br);
    nutreec[3+c*4]=br;
    /* printf("%i %i %i\n", r, p, l); */
    /* printf("%i %i %i\n", c, (int) nutreec[c*4], (int) nutreec[1+c*4]); */
  }
  
  /* finish processing the root join */
  nutreec[1+c*4]=l; nutreec[3+c*4] += bl;
  /* printf("%i %i %i\n\n",  c, (int) nutreec[c*4], (int) nutreec[1+c*4]); */
  nc=c;
  c--;

  /* Start from the root of nutreec and relabel all joins so that they
   * don't conflict with the newly created ones. */
  j=ntaxa-2;
  while(j >= 0){
    kup=(j>nc)?1:2; /* kup set so that newly created joins are not changed */
    for(k = 0; k < kup; k++){
      r=nutreec[k+j*4];
      if (r >= ntaxa){
	/* printf("c j %i %i\n",c,j); */
	for(l = 0; l < 4; l++) 
	  nutreec[l+c*4]=utreec[l+(r-ntaxa)*4]; 
	nutreec[k+j*4]=c+ntaxa;
	c--;
      }
    }
    j--;
  }
}
int sp_char(char c){
  return((c == '(' || c == ')' || c == ',' || c == ':' || c == ';'));
}
void ch_name(int lstring, char *tstring, int i){ 
  /* change name given by the first lstring chars of tstring to integer label i */
  char label[3];
  int tstl, llen;
  
  /* shorten the string and get rid of the name */
  tstl = 0;
  while(tstring[tstl+lstring] != '\0'){
    tstring[tstl] = tstring[tstl+lstring];
    tstl++;
  }
  tstring[tstl] = '\0';
  
  sprintf(label, "%i", i); /* i to char label */
  llen = strlen(label);   /* number of digits */
  
  /* make room for integer label; more efficient: combine this and shorten above */
  while(tstl >= 0){
    tstring[tstl+llen] = tstring[tstl];
    tstl--;
  }
  /* place the integer label where name used to be */
  for(tstl = 0; tstl < llen; tstl++)
    tstring[tstl] = label[tstl];
}
int has_bl(char *tstring)
{
  while(*tstring != ';')
    if(*tstring == ':') /* signifies start of bl */
      return(1);
    else
      tstring++;
  return(0); /* must not be bl if we got this far */
}
void ch_labels_nbl(int ntaxa, char *tstring, char names[][11]){
  /* change all names in tstring to integer labels (no branch lengths)*/
  char taxname[11];
  int ltstring, i;

  while(*tstring != ';'){
    while(*tstring == ')' || *tstring == '(' || *tstring == ','){
      tstring++;
    }
    if(*tstring == ';') break;  /* end of tstring, no more names */

    ltstring = 0; /* beginning of new name */
    while(*tstring != ')' && *tstring != '(' && *tstring != ','){
      taxname[ltstring] = *tstring;
      tstring++;
      ltstring++;
    }

    for(i=ltstring; i < 10; i++) taxname[i] = ' '; /* pad w/ blanks */
    taxname[10] = '\0'; /* name is in taxname */
    for(i = 0; i < ltstring; i++) tstring--; /* go to start of name */
    
    for(i = 0; i < ntaxa; i++){            /* check which taxa it is */
      if (strncmp(taxname, names[i], 10) == 0){ /* i is label for taxa */
	ch_name(ltstring, tstring, i);	  
	break;
      }
    }
    /* go to end of name */
    while(*tstring != '(' && *tstring != ')' && *tstring != ',') tstring++;
  }
}
void ch_labels(int ntaxa, char *tstring, char names[][11]){
  /* change all names in tstring to integer labels */
  char taxname[11];
  int lstring, i;
  
  while(*tstring != ';'){
    while(*tstring != ':' && *tstring != ';'){ /* name always before : */
      tstring++;
    }
    if (*tstring == ';') break;  /* end of tstring, no more names */
    tstring--; lstring = 0;      /* end of a possible name */
    while(!sp_char(*tstring)){   /* find the beginning */
      tstring--; lstring++;
    }
    if (*tstring == '(' || *tstring == ','){ /* condition for a name */
      tstring++;
      strncpy(taxname, tstring, lstring);    /* name is in taxname */
      for(i=lstring; i < 10; i++) taxname[i] = ' '; /* pad w/ blanks */
      taxname[10] = '\0';

/*       for(i = 0; i < lstring; i++) putchar(taxname[i]); printf("\n");  */
      for(i = 0; i < ntaxa; i++){            /* check which taxa it is */
/*  	printf("%s\n", names[i]);  */
	if (strncmp(taxname, names[i], 10) == 0){ /* i is label for taxa */
	  ch_name(lstring, tstring, i);	  
	  break;
	}
      }
/*       printf("%s\n\n", tstring); */ 
    }
    while(*tstring != ':') tstring++; /* get back to : */
    tstring++; /* process the rest of string */
  }
}
int is_wsp(char c){
  if(c==' ' || c=='\t' || c=='\n' || c=='\t') return(1);
  return(0);
}
void rmblntree(char *tstring, int rwsp){
  /* I: wsp=1 => remove whitespace */
  int s=0,e=0;
  while(tstring[e]!=';'){
    if(tstring[e]==':')
      while(tstring[e]!=',' && tstring[e]!=')') e++;
    if(!rwsp || (tstring[e]!=' ' && tstring[e]!='\t'&& tstring[e]!='\r' && 
		 tstring[e]!='\n'))
      tstring[s++]=tstring[e];
    e++;
  }
  tstring[s++]=';';
  tstring[s]='\0';
}
int read_tree(FILE *fp, char *tstring)
{
  /* assumes 
   *   '[' and ']' delimit comments only
   *    '][' won't be used to end a comment and start a new one
   *   white space can be removed */
  int ltstring;
  char c;

  ltstring = 0;
  while((c = getc(fp)) != ';'){ 
    if (c != ' ' && c != '\t' && c != '\r' && c != '\n'){ /* no white space */
      if (c == '[') /* start of comment; ignored */
	while((c = getc(fp)) != ']') ;
      else
	tstring[ltstring++]=c;
    }
  } 
  tstring[ltstring++]=';';
  tstring[ltstring++]='\0';
  return(ltstring);
}
double get_dlist(char *tstr, int ss, int de, int *ssn, int *ls, int *le)
{
  /* IN:
   * tstr - tree
   * ss - char position to start search
   * de - end position of dlist to be searched
   * OUT:
   * ssn - char position to start new search
   * ls - start pos of label; (ls-1) gives end of dlist found
   * le - end pos of label; 
   * RETURN: edge length (0.1 if not present) */

  char c,els[20]; /* note the limit on edge-length string */
  int lb=0,rb=0,lrb,i,j=0;
  double el=0.1;

  i=ss;
  lrb=ss-1;
  while(i <= de-1){
    if ((c=tstr[i]) == '(') lb++;
    if (c == ')'){rb++; lrb=i;}
    if (c == ',' && lb==rb) break;
    i++;
  }
  *ssn=i+1;
  *ls=lrb+1;
  i=lrb+1; 
  while((c=tstr[i]) != ':' && c != ','&& c != ')') i++;
  *le=i-1;
  if(c == ':'){
    while((c=tstr[++i]) != ',' && c != ')') els[j++]=c;
    els[j]='\0';
    el=atof(els);
  }
  return(el);
}
void underscore2blank(char *s){
  while(*s++ != '\0')
    if(*s=='_') *s=' ';
}
char **treecs(int ntaxa, char *tstr, char names[][11], double *utreec, 
	      int has_names){
  /* IN:
   * tstr - Newick tree w/o comments, whitespace and ; (see read_tree)
   * has_names = 1 if names are available
   * IN/OUT:
   * names - if has_names, used to determine ordering 0,...,ntaxa-1 of 
   *         terminal labels
   *         o/w gives names of the taxa labeled 0,...,ntaxa-1 in utreec
   * OUT:
   * utreec 
   * RETURN:
   * ilabels - internal nodel labels; (ntaxa-2); might be uninitialized
   *    not allocated as a contiguous block
   *
   * COMMENTS:
   * Based on observation that a root subtree is either
   *    (a) a leaf and has no `(' or ')' 
   * or (b) a dlist + label and el, and has equal numbers of `(' and ')' 
   *
   * violations of Newick standard
   * - quoted labels not allowed
   * - '[' and ']' only used to delimit comments
   * - limits of 10 non-null chars on leaf names
   * - underscores are not converted to blanks
   */

  int *ds; /* starting char positions of dlists */
  int *de; /* ending char positions of dlists */
  int cd=0,td=1; /* current and total dlists */
  int cj; /* current join */
  int cil; /* current internal label */
  int ctl=0; /* current terminal label */
  int *ul; /* utreec labels */
  int *ls,*le,*is_leaf,k,i,ss,K,j,ssn,ir,il;
  char **ilabels,tname[11];
  double *el; 

  int tdo=1; /* testing */

  el = (double *) malloc((size_t) ntaxa*sizeof(double));
  ds = (int *) malloc((size_t) 2*ntaxa*sizeof(int));
  de = (int *) malloc((size_t) 2*ntaxa*sizeof(int));
  ls = (int *) malloc((size_t) ntaxa*sizeof(int));
  le = (int *) malloc((size_t) ntaxa*sizeof(int));
  is_leaf = (int *) malloc((size_t) ntaxa*sizeof(int));
  ul = (int *) malloc((size_t) ntaxa*sizeof(int));
  ilabels = (char **)malloc((size_t) (ntaxa-2)*sizeof(char *));

  /* if(has_names) */
  /*   for(i = 0; i < ntaxa; i++) underscore2blank(names[i]); */

  for(i = 0; i < ntaxa; i++) el[i]=0;
  
  *ds=0; *de=0; *is_leaf=0;
  while(tstr[*de] != '\0')(*de)++;
  while(tstr[*de] != ')') (*de)--; /* root label and edge ignored */

  cj=ntaxa-2;
  cil=2*ntaxa-3;
  while (cj >= 0){/* Main loop */

    /* obtain all root subtrees of the current dlist and update dlists */
    ss=ds[cd]+1;
    K=0;
    while(ss <= de[cd]-1){
      el[K]=get_dlist(tstr,ss,de[cd],&ssn,&ls[K],&le[K]);
      is_leaf[K]=(ls[K] == ss)?1:0;
      if (!is_leaf[K]){
	ds[td]=ss;
	de[td]=ls[K]-1;
	td++;
      }
      ss=ssn;
      K++;
    }

    /* attach ulabels to subtrees */
    for(i = 0; i < K-2; i++){
      ilabels[cil-ntaxa]=(char *)malloc(sizeof(char));
      ilabels[cil-ntaxa][0]='\0';
      cil--;
    }
    for(i=0; i < K; i++){
      if (!is_leaf[i]){
	ilabels[cil-ntaxa]=(char *)malloc((le[i]-ls[i]+2)*sizeof(char));
	memcpy(ilabels[cil-ntaxa],&tstr[ls[i]],(le[i]-ls[i]+1)*sizeof(char));
	ilabels[cil-ntaxa][le[i]+1-ls[i]]='\0';
	/* underscore2blank(ilabels[cil-ntaxa]); */
	ul[i]=cil;
	cil--;
      }
      if (is_leaf[i] && !has_names){
	memcpy(names[ctl],&tstr[ls[i]],(le[i]-ls[i]+1)*sizeof(char));
	for(k=le[i]+1-ls[i]; k < 10; k++) names[ctl][k]=' ';
	names[ctl][10]='\0';
	/* underscore2blank(names[ctl]); */
	ul[i]=ctl;
	ctl++;
      }
      if (is_leaf[i] && has_names){
	memcpy(tname,&tstr[ls[i]],(le[i]-ls[i]+1)*sizeof(char));
	for(k=le[i]+1-ls[i]; k < 10; k++) tname[k]=' ';
	tname[10]='\0';
	/* underscore2blank(tname); */
	for(j = 0; j < ntaxa; j++) 
	  if (strncmp(tname,names[j],11)==0) break;
	if(j >= ntaxa){
	  printf("treecs: unable to find %s among taxa\n",tname); exit(0);}
	ul[i]=j;
      } 
    }/* end of attach ulabels ... */
    
    /* printf("current subtree %i\n",ntaxa+cj); */
    /* for(j = ds[cd]; j <= de[cd]; j++) putchar(tstr[j]); printf("\n"); */
    /* printf("number of subtrees: %i\n",K); */
    /* for(i=0; i < K; i++){ */
    /*   printf("%i %f ",ul[i],el[i]); */
    /*   if(!is_leaf[i]){ */
    /* 	for(j = ds[tdo]; j <= de[tdo]; j++) putchar(tstr[j]); */
    /* 	tdo++; */
    /*   } */
    /*   printf("\n"); */
    /* } */
    /* tdo=td; */

    /* adjust labels of existing internal edges for multifurcation */
    if(K>2)
      for(i=cj+1; i<ntaxa-1; i++)
	for(k=0; k<2; k++) 
	  if(utreec[k+i*4]>ntaxa && utreec[k+i*4]<ntaxa+cj) utreec[k+i*4] -= K-2;

    /* add root subtrees to utreec */
    for(i = 0; i < K-2; i++){
      utreec[cj*4]=ul[i]; utreec[1+cj*4]=cj-1+ntaxa; 
      utreec[2+cj*4]=el[i]; utreec[3+cj*4]=0.0;
      cj--;
    }
    if (ul[K-2] < ul[K-1]){
      ir=K-2;il=K-1;
    }
    else{
      ir=K-1;il=K-2;
    }
    utreec[cj*4]=ul[ir]; utreec[2+cj*4]=el[ir];
    utreec[1+cj*4]=ul[il]; utreec[3+cj*4]=el[il];
    cj--;
    cd++;
    
    /* for(i=cj+1; i<ntaxa-1; i++) */
    /*   printf("%i %i %i %f %f\n",i+ntaxa,(int) utreec[i*4],(int) utreec[1+i*4], */
    /* 	     utreec[2+i*4],utreec[3+i*4]); */

  } /* end of Main loop */
/*   pr_utreec(stdout,ntaxa,utreec); */
/*   for(i = 0; i < ntaxa; i++) */
/*     printf("%s\n",names[i]); */
  
  free(is_leaf); 
  free(ul); 
  free(le); 
  free(ls); 
  free(de); 
  free(ds); 
  free(el); 
  return(ilabels);
}
void treecsnl(int ntaxa, char *tstr, char names[][11], double *utreec, 
	      int has_names){
  char **ilabels; 
  int j;
  ilabels=treecs(ntaxa, tstr, names, utreec, has_names);
  for(j=0; j<ntaxa-2; j++) free(ilabels[j]);
  free(ilabels);
}
char **read_treecs(FILE *treefile, int ntaxa, char names[][11], double *utreec, 
		   int has_names){
  /* IN:
   * treefile - currently reading a new tree
   * IN/OUT:
   * names - if has_names, used to determine ordering 0,...,ntaxa-1 of 
   *         terminal labels
   *         o/w gives names of the taxa labeled 0,...,ntaxa-1 in utreec
   * OUT:
   * utreec 
   * RETURN:
   * ilabels - internal nodel labels; (ntaxa-2); each might only have '\0'
   *    not allocated as a contiguous block
   *
   * COMMENTS:
   * Based on observation that a root subtree is either
   *    (a) a leaf and has no `(' or ')' *
   * or (b) a dlist + label and el, and has equal numbers of `(' and ')' 
   *
   * violations of Newick standard
   * - quoted labels not allowed
   * - '[' and ']' only used to delimit comments
   * - limits of 10 non-null chars on leaf names 
   * - underscores are not converted to blanks
   */
  char tstring[10000];
  char **ilabels;
  read_tree(treefile,tstring);
  /* printf("%s\n",tstring); */
  ilabels=treecs(ntaxa,tstring,names,utreec,has_names);
  return(ilabels);
}
void read_treecsnl(FILE *treefile, int ntaxa, char names[][11], double *utreec, 
		   int has_names){
  char tstring[10000];
  read_tree(treefile,tstring);
  /* printf("%s\n",tstring); */
  treecsnl(ntaxa,tstring,names,utreec,has_names);
}
void rmexspce(int numsp, char names[][11]){
  int i,j;
  for(i = 0; i < numsp; i++)
    for(j = 9; j >= 0; j--)
      if (names[i][j] != ' ' && names[i][j] != '\0' &&  names[i][j] != '\t' && names[i][j] != '\r'){
	names[i][j+1] = '\0';
	break;
      }
}
int ndig_intpart(double f){
  int d=1,ndigits=0;
  while(((int) f/d) > 0){
    d *= 10;
    ndigits++;
  }
  if (ndigits == 0) ndigits=1;
  return(ndigits);
}
char *ut2in(int numsp, double *utreec, char names[][11], char **ilabels){
  /* `excess' space removed from names */
  char **tstring,*tstr;
  int i,alen,len,r,l;
  double x;

  tstring = (char **) malloc((size_t) (2*numsp-2)*sizeof(char *));

  rmexspce(numsp,names);
  for(i = 0; i < numsp; i++){
    len = lstring(names[i])+1;
    tstring[i] = (char *) malloc((size_t) len*sizeof(char));
    strcpy(tstring[i], names[i]);
  }

  for(i = 0; i < (numsp-3); i++){
    /* memory allocation */
    r = utreec[i*4]; l = utreec[1+i*4];
    len = lstring(tstring[r]) + lstring(tstring[l]) +
      ndig_intpart(utreec[2+i*4]) + ndig_intpart(utreec[3+i*4]) + 18;
    if (r >= numsp) len += lstring(ilabels[r-numsp]);
    if (l >= numsp) len += lstring(ilabels[l-numsp]);
    tstring[i+numsp] = (char *) malloc((size_t) len*sizeof(char));
    /* create subtree */
    if (r >= numsp && l >= numsp)
      alen=sprintf(tstring[i+numsp],"(%s%s:%.5f,%s%s:%.5f)",tstring[r],
	      ilabels[r-numsp],utreec[2+i*4],tstring[l],ilabels[l-numsp],
	      utreec[3+i*4]);
    if (r >= numsp && l < numsp)
      alen=sprintf(tstring[i+numsp],"(%s%s:%.5f,%s:%.5f)",tstring[r],
	      ilabels[r-numsp],utreec[2+i*4],tstring[l],utreec[3+i*4]);
    if (r < numsp && l >= numsp)
      alen=sprintf(tstring[i+numsp],"(%s:%.5f,%s%s:%.5f)",tstring[r],
	     utreec[2+i*4],tstring[l],ilabels[l-numsp],utreec[3+i*4]);
    if (r < numsp && l < numsp)
      alen=sprintf(tstring[i+numsp],"(%s:%.5f,%s:%.5f)",tstring[r],
	      utreec[2+i*4],tstring[l],utreec[3+i*4]);
    /* printf("%i %i: %s\n",alen+1,i+numsp,tstring[i+numsp]); */
  }
  /* PHYLIP requires trifurcation */
  /* memory allocation */
  r = utreec[(numsp-3)*4]; l = utreec[1+(numsp-3)*4];
  len = lstring(tstring[r]) + lstring(tstring[l]) + 16 +
    ndig_intpart(utreec[2+(numsp-3)*4])+ndig_intpart(utreec[3+(numsp-3)*4]);
  if (r >= numsp) len += lstring(ilabels[r-numsp]);
  if (l >= numsp) len += lstring(ilabels[l-numsp]);
  tstring[2*numsp-3] = (char *) malloc((size_t) len*sizeof(char));
  /* create subtree */
  i=numsp-3;
  if (r >= numsp && l >= numsp)
    alen=sprintf(tstring[i+numsp],"%s%s:%.5f,%s%s:%.5f",tstring[r],
	    ilabels[r-numsp],utreec[2+i*4],tstring[l],ilabels[l-numsp],
	    utreec[3+i*4]);
  if (r >= numsp && l < numsp)
    alen=sprintf(tstring[i+numsp],"%s%s:%.5f,%s:%.5f",tstring[r],
	    ilabels[r-numsp],utreec[2+i*4],tstring[l],utreec[3+i*4]);
  if (r < numsp && l >= numsp)
    alen=sprintf(tstring[i+numsp],"%s:%.5f,%s%s:%.5f",tstring[r],
	    utreec[2+i*4],tstring[l],ilabels[l-numsp],utreec[3+i*4]);
  if (r < numsp && l < numsp)
    alen=sprintf(tstring[i+numsp],"%s:%.5f,%s:%.5f",tstring[r],
	    utreec[2+i*4],tstring[l],utreec[3+i*4]);
  /* printf("%i %i %i %i %i: %s\n",r,l,len,alen+1,2*numsp-3,tstring[2*numsp-3]); */
  /* memory allocation for final string */
  r = utreec[(numsp-2)*4]; l = utreec[1+(numsp-2)*4];
  r = (r < l)?r:l;
  x=utreec[2+(numsp-2)*4]+utreec[3+(numsp-2)*4];
  len = lstring(tstring[2*numsp-3]) + lstring(tstring[r]) + 12 + 
    ndig_intpart(x);
  if (r >= numsp) len += lstring(ilabels[r-numsp]);
  tstr=(char *) malloc((size_t) len*sizeof(char));
  /* tree */
  if (r >= numsp)
    alen=sprintf(tstr,"(%s,%s%s:%.5f);",tstring[2*numsp-3],tstring[r],
	    ilabels[r-numsp],x);
  if (r < numsp)
    alen=sprintf(tstr,"(%s,%s:%.5f);",tstring[2*numsp-3],tstring[r],x);
  tstr[len-1] = '\0';
  /* printf("%i %i %i %s\n",r,len,alen+1,tstr); */
  for(i = 0; i < 2*numsp-2; i++) free(tstring[i]);
  free(tstring);
  return(tstr);
}
void ut2n(int numsp, double *utreec, char names[][11], char *tstr){
  char **tstring;
  int i,len,r,l;

  tstring = (char **) malloc((size_t) (2*numsp-2)*sizeof(char *));

  for(i = 0; i < numsp; i++){
    len = lstring(names[i]);
    tstring[i] = (char *) malloc((size_t) len*sizeof(char));
    strcpy(tstring[i], names[i]);
    /* remove trailing blanks */
    while(tstring[i][len-1]==' ') len--;
    tstring[i][len]='\0';
  }

  for(i = 0; i < (numsp-3); i++){
    r = utreec[i*4]; l = utreec[1+i*4];
    len = lstring(tstring[r]) + lstring(tstring[l])+23; /* space for branch lengths in the 100s */
    tstring[i+numsp] = (char *) malloc((size_t) len*sizeof(char));
    sprintf(tstring[i+numsp],"(%s:%.5f,%s:%.5f)",tstring[r],utreec[2+i*4],tstring[l],
	    utreec[3+i*4]);
  }
  /* PHYLIP requires trifurcation */
  r = utreec[(numsp-3)*4]; l = utreec[1+(numsp-3)*4];
  len = lstring(tstring[r]) + lstring(tstring[l]) + 21;
  tstring[2*numsp-3] = (char *) malloc((size_t) len*sizeof(char));
  sprintf(tstring[2*numsp-3],"%s:%.5f,%s:%.5f",tstring[r],utreec[2+(numsp-3)*4],tstring[l],
	    utreec[3+(numsp-3)*4]);
  r = utreec[(numsp-2)*4]; l = utreec[1+(numsp-2)*4];
  r = (r < l)?r:l;
  sprintf(tstr,"(%s,%s:%.5f);",tstring[2*numsp-3],tstring[r],
	    utreec[2+(numsp-2)*4]+utreec[3+(numsp-2)*4]);
  len = lstring(tstr);
  tstr[len] = '\0';
  for(i = 0; i < 2*numsp-2; i++) free(tstring[i]);
  free(tstring);
}
void ut2nt(int numsp, double *utreec, char names[][11], char *tstr){
  char **tstring;
  int i,len,r,l;

  tstring = (char **) malloc((size_t) (2*numsp-2)*sizeof(char *));

  for(i = 0; i < numsp; i++){
    len = lstring(names[i]);
    tstring[i] = (char *) malloc((size_t) len*sizeof(char));
    strcpy(tstring[i], names[i]);
    /* remove trailing blanks */
    while(tstring[i][len-1]==' ') len--;
    tstring[i][len]='\0';
  }

  for(i = 0; i < (numsp-3); i++){
    r = utreec[i*4]; l = utreec[1+i*4];
    len = lstring(tstring[r]) + lstring(tstring[l])+4; 
    tstring[i+numsp] = (char *) malloc((size_t) len*sizeof(char));
    sprintf(tstring[i+numsp],"(%s,%s)",tstring[r],tstring[l]);
  }
  /* PHYLIP requires trifurcation */
  r = utreec[(numsp-3)*4]; l = utreec[1+(numsp-3)*4];
  len = lstring(tstring[r]) + lstring(tstring[l]) +2;
  tstring[2*numsp-3] = (char *) malloc((size_t) len*sizeof(char));
  sprintf(tstring[2*numsp-3],"%s,%s",tstring[r],tstring[l]);
  r = utreec[(numsp-2)*4]; l = utreec[1+(numsp-2)*4];
  r = (r < l)?r:l;
  sprintf(tstr,"(%s,%s);",tstring[2*numsp-3],tstring[r]);
  len = lstring(tstr);
  tstr[len] = '\0';
  for(i = 0; i < 2*numsp-2; i++) free(tstring[i]);
  free(tstring);
}
int icomp(const int *a, const int *b) 
{
  return((*a <= *b)?-1:1);
}
void isort(int n, int *x)
{
  qsort(x, n, (size_t) sizeof(int), (int (*)(const void *, const void *)) icomp);
}
double dot(int n, double *u, int incu, double *v, int incv)
{
  double f;

  for(f = 0.0; n-- > 0; u += incu, v += incv)
    f += (*u) * (*v);

  return(f);
}
int chol(int m, double *l)
{
  int i = 0, ii = 0, j, ij, ik, im;

  for(i = 0, ii = 0, im = 0; i < m; i++, ii += (m+1), im += m){
    for(j = 0, ij = i;  j < i;  j++, ij += m)
      l[ij] = (l[ij] - dot(j, l+i, m, l+j, m))/l[j + j*m];
    for(ik = i; ik < im; ik += m) 
      l[ii] -= l[ik]*l[ik];
    if (l[ii] <= 0.0) 
      return(1);
    l[ii] = sqrt(l[ii]);
  }
  return(0);
}
void cholf(int m, double *l, double *b)
{
  int i, ii;

  for(i = 0; i < m; i++)
    b[i] = (b[i] - dot(i, l+i, m, b, 1))/l[i + i*m];
  for(i = m - 1; i >= 0; i--){
    ii = i + i*m;
    b[i] = (b[i] - dot(m-i-1, l+ii+1, 1, b+i+1, 1))/l[ii];
  }
}
void cholh(int m, double *l, double *b)
{
  int i;
  /* solution of l x = b */
  for(i = 0; i < m; i++)
    b[i] = (b[i] - dot(i, l+i, m, b, 1))/l[i + i*m];
}
int choltr(int m, double *l){
  int i,j,k,ij,ik,jk;

  for(ij=0, i=0; i<m; i++, ij++){
    for(j=0; j<i; j++, ij++){
      ik=(i*(i+1))/2;
      jk=(j*(j+1))/2;
      for(k=0; k<j; k++, ik++, jk++) l[ij] -= l[ik]*l[jk];
      l[ij] /= l[jk]; /* jk = jj */
    }
    ik=(i*(i+1))/2;
    for(k=0; k<i; k++, ik++) l[ij] -= l[ik]*l[ik];
    if(l[ij] <= 0.0) return(1);
    l[ij] = sqrt(l[ij]);
  }
  
  return(0);
}
int lstring(char * string){
  /* returns length of string */
  int l = 0;
  while (string[l] != '\0') l++;
  return(l);
}
int eigenRealSym(double A[], int n, double Root[], double Offdiag[])
{
/* This finds the eigen solution of a real symmetrical matrix A[n*n].  In return, 
   A has the right vectors and Root has the eigenvalues. work[n] is the working space.
   The matrix is first reduced to a tridiagonal matrix using HouseholderRealSym(), 
   and then using the QL algorithm with implicit shifts.  

   Adapted from routine tqli in Numerical Recipes in C, with reference to LAPACK
   Ziheng Yang, 23 May 2001
*/
   int status=0;
   HouseholderRealSym(A, n, Root, Offdiag);
   status=EigenTridagQLImplicit(Root, Offdiag, n, A);
   EigenSort(Root, A, n);

   return(status);
}
void EigenSort(double d[], double U[], int n)
{
/* this sorts the eigen values d[] and rearrange the (right) eigen vectors U[]
*/
   int k,j,i;
   double p;

   for (i=0;i<n-1;i++) {
      p=d[k=i];
      for (j=i+1;j<n;j++)
         if (d[j] >= p) p=d[k=j];
      if (k != i) {
         d[k]=d[i];
         d[i]=p;
         for (j=0;j<n;j++) {
            p=U[j*n+i];
            U[j*n+i]=U[j*n+k];
            U[j*n+k]=p;
         }
      }
   }
}
void HouseholderRealSym(double a[], int n, double d[], double e[])
{
/* This uses HouseholderRealSym transformation to reduce a real symmetrical matrix 
   a[n*n] into a tridiagonal matrix represented by d and e.
   d[] is the diagonal (eigends), and e[] the off-diagonal.
*/
   int m,k,j,i;
   double scale,hh,h,g,f;

   for (i=n-1;i>=1;i--) {
      m=i-1;
      h=scale=0;
      if (m > 0) {
         for (k=0;k<=m;k++)
            scale += fabs(a[i*n+k]);
         if (scale == 0)
            e[i]=a[i*n+m];
         else {
            for (k=0;k<=m;k++) {
               a[i*n+k] /= scale;
               h += a[i*n+k]*a[i*n+k];
            }
            f=a[i*n+m];
            g=(f >= 0 ? -sqrt(h) : sqrt(h));
            e[i]=scale*g;
            h -= f*g;
            a[i*n+m]=f-g;
            f=0;
            for (j=0;j<=m;j++) {
               a[j*n+i]=a[i*n+j]/h;
               g=0;
               for (k=0;k<=j;k++)
                  g += a[j*n+k]*a[i*n+k];
               for (k=j+1;k<=m;k++)
                  g += a[k*n+j]*a[i*n+k];
               e[j]=g/h;
               f += e[j]*a[i*n+j];
            }
            hh=f/(h*2);
            for (j=0;j<=m;j++) {
               f=a[i*n+j];
               e[j]=g=e[j]-hh*f;
               for (k=0;k<=j;k++)
                  a[j*n+k] -= (f*e[k]+g*a[i*n+k]);
            }
         }
      } 
      else
         e[i]=a[i*n+m];
      d[i]=h;
   }
   d[0]=e[0]=0;

   /* Get eigenvectors */
   for (i=0;i<n;i++) {
      m=i-1;
      if (d[i]) {
         for (j=0;j<=m;j++) {
            g=0;
            for (k=0;k<=m;k++)
               g += a[i*n+k]*a[k*n+j];
            for (k=0;k<=m;k++)
               a[k*n+j] -= g*a[k*n+i];
         }
      }
      d[i]=a[i*n+i];
      a[i*n+i]=1;
      for (j=0;j<=m;j++) a[j*n+i]=a[i*n+j]=0;
   }
}
int EigenTridagQLImplicit(double d[], double e[], int n, double z[])
{
/* This finds the eigen solution of a tridiagonal matrix represented by d and e.  
   d[] is the diagonal (eigenvalues), e[] is the off-diagonal
   z[n*n]: as input should have the identity matrix to get the eigen solution of the 
   tridiagonal matrix, or the output from HouseholderRealSym() to get the 
   eigen solution to the original real symmetric matrix.
   z[n*n]: has the orthogonal matrix as output

   Adapted from routine tqli in Numerical Recipes in C, with reference to
   LAPACK fortran code.
   Ziheng Yang, May 2001
*/
   int m,j,iter,niter=30, status=0, i,k;
   double s,r,p,g,f,dd,c,b, aa,bb;

   for (i=1;i<n;i++) e[i-1]=e[i];  e[n-1]=0;
   for (j=0;j<n;j++) {
      iter=0;
      do {
         for (m=j;m<n-1;m++) {
            dd=fabs(d[m])+fabs(d[m+1]);
            if (fabs(e[m])+dd == dd) break;  /* ??? */
         }
         if (m != j) {
            if (iter++ == niter) {
               status=-1;
               break;
            }
            g=(d[j+1]-d[j])/(2*e[j]);

            /* r=pythag(g,1); */

            if((aa=fabs(g))>1)  r=aa*sqrt(1+1/(g*g));
            else                r=sqrt(1+g*g);

            g=d[m]-d[j]+e[j]/(g+SIGN(r,g));
            s=c=1;
            p=0;
            for (i=m-1;i>=j;i--) {
               f=s*e[i];
               b=c*e[i];

               /*  r=pythag(f,g);  */
               aa=fabs(f); bb=fabs(g);
               if(aa>bb)       { bb/=aa;  r=aa*sqrt(1+bb*bb); }
               else if(bb==0)             r=0;
               else            { aa/=bb;  r=bb*sqrt(1+aa*aa); }

               e[i+1]=r;
               if (r == 0) {
                  d[i+1] -= p;
                  e[m]=0;
                  break;
               }
               s=f/r;
               c=g/r;
               g=d[i+1]-p;
               r=(d[i]-g)*s+2*c*b;
               d[i+1]=g+(p=s*r);
               g=c*r-b;
               for (k=0;k<n;k++) {
                  f=z[k*n+i+1];
                  z[k*n+i+1]=s*z[k*n+i]+c*f;
                  z[k*n+i]=c*z[k*n+i]-s*f;
               }
            }
            if (r == 0 && i >= j) continue;
            d[j]-=p; e[j]=g; e[m]=0;
         }
      } while (m != j);
   }
   return(status);
}
int DiscreteGamma (double freqK[], double rK[], double alpha, double beta, int K, int dgammamean)
{
/* discretization of gamma distribution with equal proportions in each 
   category.
*/
   int i;
   double t, factor=alpha/beta*K, lnga1;

   if (dgammamean==0) {
      lnga1=LnGamma(alpha+1);
      for (i=0; i<K-1; i++) /* cutting points, Eq. 9 */
         freqK[i]=InverseCDFGamma((i+1.0)/K, alpha, beta);
      for (i=0; i<K-1; i++) /* Eq. 10 */
         freqK[i]=IncompleteGamma(freqK[i]*beta, alpha+1, lnga1);

      for(i=0; i<K; i++) rK[i]=freqK[i];

      rK[0] = freqK[0]*factor;
      rK[K-1] = (1-freqK[K-2])*factor;
      for (i=1; i<K-1; i++)  rK[i] = (freqK[i]-freqK[i-1])*factor;
   }
   else {
      for(i=0; i<K; i++) rK[i]=InverseCDFGamma((i*2.+1)/(2.*K), alpha, beta);
      for(i=0,t=0; i<K; i++) t+=rK[i];
      for(i=0; i<K; i++) rK[i]*=factor/t;
   }
   for (i=0; i<K; i++) freqK[i]=1.0/K;

   return (0);
}
void error2 (char * message)
{ printf("\nError: %s.\n", message); exit(-1); }
long factorial (int n)
{
   long f=1, i;
   if (n>10) error2("n>10 in factorial");
   for (i=2; i<=(long)n; i++) f *= i;
   return (f);
}
double LnGamma (double x)
{
/* returns ln(gamma(x)) for x>0, accurate to 10 decimal places.
   Stirling's formula is used for the central polynomial part of the procedure.

   Pike MC & Hill ID (1966) Algorithm 291: Logarithm of the gamma function.
   Communications of the Association for Computing Machinery, 9:684
*/
   double f=0, fneg=0, z, lng;
   int nx=(int)x;

   if((double)nx==x && nx>1 && nx<12)
      lng=log((double)factorial(nx-1));
   else {
      if(x<=0) {
         error2("LnGamma not implemented for x<0");
         if((int)x-x==0) { puts("lnGamma undefined"); return(-1); }
         for (fneg=1; x<0; x++) fneg/=x;
         if(fneg<0) error2("strange!! check lngamma");
         fneg=log(fneg);
      }
      if (x<7) {
         f=1;  z=x-1;
         while (++z<7)  f*=z;
         x=z;   f=-log(f);
      }
      z = 1/(x*x);
      lng = fneg+ f + (x-0.5)*log(x) - x + .918938533204673 
             + (((-.000595238095238*z+.000793650793651)*z-.002777777777778)*z
                  +.083333333333333)/x;
   }
   return  lng;
}
double InverseCDFNormal (double prob)
{
/* returns z so that Prob{x<z}=prob where x ~ N(0,1) and (1e-12)<prob<1-(1e-12)
   returns (-9999) if in error
   Odeh RE & Evans JO (1974) The percentage points of the normal distribution.
   Applied Statistics 22: 96-97 (AS70)

   Newer methods:
     Wichura MJ (1988) Algorithm AS 241: the percentage points of the
       normal distribution.  37: 477-484.
     Beasley JD & Springer SG  (1977).  Algorithm AS 111: the percentage 
       points of the normal distribution.  26: 118-121.
*/
   double a0=-.322232431088, a1=-1, a2=-.342242088547, a3=-.0204231210245;
   double a4=-.453642210148e-4, b0=.0993484626060, b1=.588581570495;
   double b2=.531103462366, b3=.103537752850, b4=.0038560700634;
   double y, z=0, p=prob, p1;

   p1 = (p<0.5 ? p : 1-p);
   if (p1<1e-20) z=999;
   else {
      y = sqrt (log(1/(p1*p1)));   
      z = y + ((((y*a4+a3)*y+a2)*y+a1)*y+a0) / ((((y*b4+b3)*y+b2)*y+b1)*y+b0);
   }
   return (p<0.5 ? -z : z);
}
double InverseCDFChi2 (double prob, double v)
{
/* returns z so that Prob{x<z}=prob where x is Chi2 distributed with df=v
   returns -1 if in error.   0.000002<prob<0.999998
   RATNEST FORTRAN by
       Best DJ & Roberts DE (1975) The percentage points of the 
       Chi2 distribution.  Applied Statistics 24: 385-388.  (AS91)
   Converted into C by Ziheng Yang, Oct. 1993.
*/
   double e=.5e-6, aa=.6931471805, p=prob, g, small=1e-6;
   double xx, c, ch, a=0,q=0,p1=0,p2=0,t=0,x=0,b=0,s1,s2,s3,s4,s5,s6;

   if (p<small)   return(0);
   if (p>1-small) return(9999);
   if (v<=0)      return (-1);

   g = LnGamma (v/2);
   xx=v/2;   c=xx-1;
   if (v >= -1.24*log(p)) goto l1;

   ch=pow((p*xx*exp(g+xx*aa)), 1/xx);
   if (ch-e<0) return (ch);
   goto l4;
l1:
   if (v>.32) goto l3;
   ch=0.4;   a=log(1-p);
l2:
   q=ch;  p1=1+ch*(4.67+ch);  p2=ch*(6.73+ch*(6.66+ch));
   t=-0.5+(4.67+2*ch)/p1 - (6.73+ch*(13.32+3*ch))/p2;
   ch-=(1-exp(a+g+.5*ch+c*aa)*p2/p1)/t;
   if (fabs(q/ch-1)-.01 <= 0) goto l4;
   else                       goto l2;
  
l3: 
   x = InverseCDFNormal(p);
   p1 = 0.222222/v;
   ch = v*pow((x*sqrt(p1)+1-p1), 3.0);
   if (ch>2.2*v+6)
      ch = -2*(log(1-p)-c*log(.5*ch)+g);
l4:
   q=ch;   p1=.5*ch;
   if ((t=IncompleteGamma (p1, xx, g))<0)
      error2 ("\nIncompleteGamma");
   p2=p-t;
   t=p2*exp(xx*aa+g+p1-c*log(ch));   
   b=t/ch;  a=0.5*t-b*c;

   s1=(210+a*(140+a*(105+a*(84+a*(70+60*a))))) / 420;
   s2=(420+a*(735+a*(966+a*(1141+1278*a))))/2520;
   s3=(210+a*(462+a*(707+932*a)))/2520;
   s4=(252+a*(672+1182*a)+c*(294+a*(889+1740*a)))/5040;
   s5=(84+264*a+c*(175+606*a))/2520;
   s6=(120+c*(346+127*c))/5040;
   ch+=t*(1+0.5*t*s1-b*c*(s1-b*(s2-b*(s3-b*(s4-b*(s5-b*s6))))));
   if (fabs(q/ch-1) > e) goto l4;

   return (ch);
}
double IncompleteGamma (double x, double alpha, double ln_gamma_alpha)
{
/* returns the incomplete gamma ratio I(x,alpha) where x is the upper 
           limit of the integration and alpha is the shape parameter.
   returns (-1) if in error
   ln_gamma_alpha = ln(Gamma(alpha)), is almost redundant.
   (1) series expansion,     if (alpha>x || x<=1)
   (2) continued fraction,   otherwise
   RATNEST FORTRAN by
   Bhattacharjee GP (1970) The incomplete gamma integral.  Applied Statistics,
   19: 285-287 (AS32)
*/
   int i;
   double p=alpha, g=ln_gamma_alpha;
   double accurate=1e-10, overflow=1e60;
   double factor, gin=0, rn=0, a=0,b=0,an=0,dif=0, term=0, pn[6];

   if (x==0) return (0);
   if (x<0 || p<=0) return (-1);

   factor=exp(p*log(x)-x-g);   
   if (x>1 && x>=p) goto l30;
   /* (1) series expansion */
   gin=1;  term=1;  rn=p;
 l20:
   rn++;
   term *= x/rn;   gin += term;
   if (term > accurate) goto l20;
   gin *= factor/p;
   goto l50;
 l30:
   /* (2) continued fraction */
   a = 1-p;   b = a+x+1;  term = 0;
   pn[0] = 1;  pn[1] = x;  pn[2] = x+1;  pn[3] = x*b;
   gin = pn[2]/pn[3];
 l32:
   a++;  
   b += 2;
   term++;
   an = a*term;
   for (i=0; i<2; i++) 
      pn[i+4] = b*pn[i+2] - an*pn[i];
   if (pn[5] == 0) goto l35;
   rn = pn[4]/pn[5];
   dif = fabs(gin-rn);
   if (dif > accurate) goto l34;
   if (dif <= accurate*rn) goto l42;
 l34:
   gin = rn;
 l35:
   for (i=0; i<4; i++) pn[i] = pn[i+2];
   if (fabs(pn[4]) < overflow) goto l32;
   for (i=0; i<4; i++) pn[i] /= overflow;
   goto l32;
 l42:
   gin = 1-factor*gin;

 l50:
   return (gin);
}
